#  Copyright (c) 2023-24 Higher Bar AI, PBC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Core classes for instrument evaluation engine."""

from langchain_openai import ChatOpenAI, AzureChatOpenAI
from langchain_core.prompts import SystemMessagePromptTemplate, HumanMessagePromptTemplate, ChatPromptTemplate
from langchain.memory import ConversationSummaryBufferMemory, ChatMessageHistory
from langchain.chains import ConversationChain
import re
import tiktoken
import json
import asyncio
import logging


class EvaluationEngine:
    """Main class for instrument evaluation engine."""

    summarize_model: str
    summarize_provider: str
    evaluation_model: str
    evaluation_provider: str
    openai_api_key: str
    azure_api_key: str
    azure_api_base: str
    azure_api_version: str
    tiktoken_model_name: str
    tokenizer: tiktoken.Encoding
    temperature: float
    max_retries: int
    logger: logging.Logger
    extra_evaluation_instructions: str

    def __init__(self, summarize_model: str, summarize_provider: str, evaluation_model: str, evaluation_provider: str,
                 openai_api_key: str, azure_api_key: str = "", azure_api_base: str = "", azure_api_version: str = "",
                 temperature: float = 0.1, tiktoken_model_name: str = "", max_retries: int = 3,
                 logger: logging.Logger = None, extra_evaluation_instructions: str = ""):
        """
        Initialize evaluation engine.

        :param summarize_model: LLM model to use for summarizing multistep conversations.
        :type summarize_model: str
        :param summarize_provider: Provider name for the summarization model ("openai" or "azure").
        :type summarize_provider: str
        :param evaluation_model: LLM model to use for instrument evaluation (when using Azure, the
            deployment or engine name must be the same as the model name).
        :type evaluation_model: str
        :param evaluation_provider: Provider name for the evaluation model ("openai" or "azure").
        :type evaluation_provider: str
        :param openai_api_key: API key for OpenAI services (if evaluation_provider is "openai").
        :type openai_api_key: str
        :param azure_api_key: API key for Azure services (if evaluation_provider is "azure").
        :type azure_api_key: str
        :param azure_api_base: Base URL for Azure API (if evaluation_provider is "azure").
        :type azure_api_base: str
        :param azure_api_version: Version of the Azure API (if evaluation_provider is "azure").
        :type azure_api_version: str
        :param temperature: Temperature setting for AI model responses.
        :type temperature: float
        :param tiktoken_model_name: Name of the model used with TikToken, if different from evaluation_model.
        :type tiktoken_model_name: str
        :param logger: Logger instance for logging messages.
        :type logger: logging.Logger
        :param extra_evaluation_instructions: Extra evaluation instructions (if any).
        :type extra_evaluation_instructions: str
        """

        # initialize object from constructor parameters
        self.summarize_model = summarize_model
        self.summarize_provider = summarize_provider
        self.evaluation_model = evaluation_model
        self.evaluation_provider = evaluation_provider
        self.openai_api_key = openai_api_key
        self.azure_api_key = azure_api_key
        self.azure_api_base = azure_api_base
        self.azure_api_version = azure_api_version
        self.temperature = temperature
        if not tiktoken_model_name:
            self.tiktoken_model_name = evaluation_model
        else:
            self.tiktoken_model_name = tiktoken_model_name
        self.tokenizer = tiktoken.encoding_for_model(self.tiktoken_model_name)
        self.max_retries = max_retries
        if not logger:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger
        self.extra_evaluation_instructions = extra_evaluation_instructions

    def tiktoken_len(self, text: str) -> int:
        """
        Count the number of tokens in a text string.

        :param text: Text to count tokens in.
        :type text: str
        :return: Count of tokens in text.
        :rtype: int
        """

        # convert to tokens and return count
        tokens = self.tokenizer.encode(
            text,
            disallowed_special=()
        )
        return len(tokens)

    def trim_string(self, s: str, max_tokens: int) -> str:
        """
        Trim string as needed to stay within token limit.

        :param s: String to trim.
        :type s: str
        :param max_tokens: Maximum number of tokens to allow in string.
        :type max_tokens: int
        :return: Trimmed string.
        :rtype: str
        """

        # see if the string is over the limit
        tokens = list(self.tokenizer.encode(s))
        if len(tokens) > max_tokens:
            # if so, truncate the tokens list and convert it back to a string
            tokens = tokens[:max_tokens]
            s = self.tokenizer.decode(tokens)

        return s

    @staticmethod
    def trim_json(json_str: str) -> str:
        """
        Trim common leading and trailing characters from JSON string.

        :param json_str: JSON string to trim
        :type json_str: str
        :return: Trimmed JSON string
        :rtype: str
        """

        # trim leading and trailing whitespace
        trimmed_json = json_str.strip()
        # trim enclosing code blocks
        trimmed_json = re.sub(r'^```json\n', '', trimmed_json)
        trimmed_json = re.sub(r'^```\n', '', trimmed_json)
        trimmed_json = re.sub(r'\n```$', '', trimmed_json)

        return trimmed_json

    @staticmethod
    def clean_whitespace(s: str) -> str:
        """
        Strip whitespace from prompt string in order to economize on tokens.

        :param s: Prompt string to clean.
        :type s: str
        :return: Cleaned prompt string.
        :rtype: str
        """

        # convert tabs to spaces
        s = s.replace('\t', ' ')

        # convert multiple spaces to one
        s = re.sub(' +', ' ', s)

        # remove the space from lines that only include a space
        s = re.sub('\n \n', '\n\n', s)
        s = re.sub('^ $', '', s, flags=re.MULTILINE)

        return s

    def get_chain(self, system_prompt: str = "", max_history_tokens: int = 2000,
                  starting_chat_history: list[tuple] = None) -> ConversationChain:
        """
        Get a conversation chain for use in evaluating an instrument.

        :param system_prompt: System prompt template to use for the conversation chain.
        :type system_prompt: str
        :param max_history_tokens: Maximum number of tokens to allow in the conversation history.
        :type max_history_tokens: int
        :param starting_chat_history: Starting chat history to use for the conversation chain (or None for none).
        :type starting_chat_history: list[tuple]
        :return: Conversation chain to use for instrument evaluation.
        :rtype: ConversationChain
        """

        # initialize summarization provider
        if self.summarize_provider == "azure":
            # (for this Azure implementation, the engine name always has to match the model)
            summarize_llm = AzureChatOpenAI(
                temperature=self.temperature,
                verbose=False,
                model_name=self.summarize_model,
                azure_endpoint=self.azure_api_base,
                openai_api_version=self.azure_api_version,
                deployment_name=self.summarize_model,
                openai_api_key=self.azure_api_key,
                openai_api_type="azure",
                tiktoken_model_name=self.tiktoken_model_name
            )
        else:
            summarize_llm = ChatOpenAI(
                temperature=self.temperature,
                verbose=False,
                model_name=self.summarize_model,
                openai_api_key=self.openai_api_key,
                tiktoken_model_name=self.tiktoken_model_name
            )

        # initialize evaluation provider
        if self.evaluation_provider == "azure":
            # for this Azure implementation, the engine name always has to match the model
            conversation_llm = AzureChatOpenAI(
                temperature=self.temperature,
                verbose=False,
                model_name=self.evaluation_model,
                azure_endpoint=self.azure_api_base,
                openai_api_version=self.azure_api_version,
                deployment_name=self.evaluation_model,
                openai_api_key=self.azure_api_key,
                openai_api_type="azure",
                tiktoken_model_name=self.tiktoken_model_name
            )
        else:
            conversation_llm = ChatOpenAI(
                verbose=False,
                temperature=self.temperature,
                model_name=self.evaluation_model,
                openai_api_key=self.openai_api_key,
                tiktoken_model_name=self.tiktoken_model_name
            )

        # construct starting history
        history = ChatMessageHistory()
        if starting_chat_history is not None:
            for (human_message, ai_message) in starting_chat_history:
                history.add_user_message(human_message)
                history.add_ai_message(ai_message)

        # requirement for ConversationChain history to work the way we want
        return_messages_preference = False

        # construct memory
        memory = ConversationSummaryBufferMemory(
            llm=summarize_llm,
            chat_memory=history,
            max_token_limit=max_history_tokens,
            output_key='answer',
            memory_key='chat_history',
            ai_prefix="Assistant",
            human_prefix="Human",
            return_messages=return_messages_preference,
            tiktoken_model_name=self.tiktoken_model_name)

        # prune starting history (if any)
        if starting_chat_history is not None:
            memory.prune()

        # create template for chat prompt, including history
        human_template = """Current conversation:
{chat_history}

Human: {question}
Assistant:"""

        # assemble prompts
        system_message_prompt = SystemMessagePromptTemplate.from_template(system_prompt)
        human_message_prompt = HumanMessagePromptTemplate.from_template(human_template)
        chat_prompt = ChatPromptTemplate.from_messages([system_message_prompt, human_message_prompt])

        # create and return conversation chain
        llm_chain = ConversationChain(
            llm=conversation_llm, prompt=chat_prompt, memory=memory, input_key="question", output_key="answer"
        )
        return llm_chain

    async def a_run_evaluation_chain(self, task_system_prompt: str, question: str, followups: list[dict],
                                     chat_history: list = None) -> dict:
        """
        Run an evaluation chain (asynchronously).

        :param task_system_prompt: System prompt template to use for the evaluation chain. This can include the
            {survey_context} and {survey_locations} variables to include information about the survey context. It
            should specify a specific JSON format for all responses.
        :type task_system_prompt: str
        :param question: Initial question to ask, to begin the evaluation chain. This should include the
            {survey_excerpt} variable to include the appropriate excerpt being evaluated.
        :type question: str
        :param followups: List of follow-up questions to ask, based on the JSON response to the initial question.
            Should be a list of dicts, each with a value for each of the following keys:
            "condition_func": function that returns True when the follow-up should be asked and False when it
            shouldn't (should take the following parameters: response_dict: dict, condition_key: str,
            condition_value: value to check against the one in the response dict);
            "condition_key": the key in the response dict to check;
            "condition_value": the value to check against the one in the response dict;
            "prompt_template": the template for the follow-up question to ask (which can include variables from
            the response dict).
        :type followups: list[dict]
        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        # initialize new evaluation chain
        llm_chain = self.get_chain(system_prompt=task_system_prompt, starting_chat_history=chat_history,
                                   max_history_tokens=6000)

        # initialize results
        result_dict = {"result": None, "error": None, "response": None, "history": None}
        response_dict = None
        full_history = []

        # ask our question to the LLM, retrying the appropriate number of times
        attempt = 0
        while True:
            attempt += 1
            try:
                # ask our question and record the result
                result = await llm_chain.ainvoke({"question": question})
                if chat_history is not None:
                    chat_history.append((question, result["answer"]))
                break
            except Exception as e:
                self.logger.error(f"Error occurred asking question (attempt {attempt}): {str(e)}")
                if attempt < self.max_retries:
                    # wait before retrying
                    await asyncio.sleep(5)
                else:
                    # maximum attempts reached, return error
                    result_dict["result"] = "error"
                    result_dict["error"] = f"Max retries reached on question. Last error: {str(e)}"
                    return result_dict

        # get prompt and response, record in history
        prompt = result["question"].strip()
        json_response = result["answer"].strip()
        full_history.append([prompt, json_response])

        # parse result as JSON
        try:
            response_dict = json.loads(EvaluationEngine.trim_json(json_response))
        except Exception as e:
            error_message = f"Error occurred parsing LLM's JSON response ({str(e)}); raw JSON: {json_response}"
            self.logger.error(error_message)
            result_dict["result"] = "error"
            result_dict["error"] = error_message
        else:
            # unless something goes wrong later, presume we're successful
            result_dict["result"] = "success"

            # ask follow-up questions
            for followup in followups:
                followup_result = await self.a_followup_question(**followup, response_dict=response_dict,
                                                                 llm_chain=llm_chain, chat_history=chat_history)

                # if we actually asked a follow-up, include it in the history
                if followup_result["result"] != "skipped":
                    full_history.append([prompt, json_response])
                if followup_result["result"] == "success" and followup_result["response"]:
                    # if we actually got a parsed response, update the response dict
                    response_dict.update(followup_result["response"])
                elif followup_result["result"] == "error":
                    # otherwise, if we got an error, update the result dict to reflect that
                    result_dict["error"] = followup_result["error"]
                    result_dict["result"] = "error"

        # assemble and return results
        result_dict["response"] = response_dict
        result_dict["history"] = full_history
        return result_dict

    def run_evaluation_chain(self, task_system_prompt: str, question: str, followups: list[dict],
                             chat_history: list = None) -> dict:
        """
        Run an evaluation chain (synchronously).

        :param task_system_prompt: System prompt template to use for the evaluation chain. This can include the
            {survey_context} and {survey_locations} variables to include information about the survey context. It
            should specify a specific JSON format for all responses.
        :type task_system_prompt: str
        :param question: Initial question to ask, to begin the evaluation chain. This should include the
            {survey_excerpt} variable to include the appropriate excerpt being evaluated.
        :type question: str
        :param followups: List of follow-up questions to ask, based on the JSON response to the initial question.
            Should be a list of dicts, each with a value for each of the following keys:
            "condition_func": function that returns True when the follow-up should be asked and False when it
            shouldn't (should take the following parameters: response_dict: dict, condition_key: str,
            condition_value: value to check against the one in the response dict);
            "condition_key": the key in the response dict to check;
            "condition_value": the value to check against the one in the response dict;
            "prompt_template": the template for the follow-up question to ask (which can include variables from
            the response dict).
        :type followups: list[dict]
        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        # run asynchronous process synchronously
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.a_run_evaluation_chain(
                task_system_prompt=task_system_prompt,
                question=question,
                followups=followups,
                chat_history=chat_history
            )
        )

    async def a_followup_question(self, condition_func: callable, condition_key: str, condition_value,
                                  prompt_template: str, response_dict: dict, llm_chain: ConversationChain,
                                  chat_history: list = None) -> dict:
        """
        Ask a follow-up question (asynchronously).

        :param condition_func: Function to call to evaluate whether the follow-up should be asked (True) or not (False).
        :type condition_func: callable
        :param condition_key: Key to check in the response dictionary.
        :type condition_key: str
        :param condition_value: Value to check for in the response dictionary, according to the logic of condition_func.
        :type condition_value: Any
        :param prompt_template: Template for the follow-up question to ask (which can include variables from the
            response dict).
        :type prompt_template: str
        :param response_dict: Full response dictionary.
        :type response_dict: dict
        :param llm_chain: Conversation chain to use for asking the follow-up question.
        :type llm_chain: ConversationChain
        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :return: A dict with result ("success", "error", or "skipped"), error (if result is "error"), prompt (a str),
            response_json (a str), and response (a dict).
        :rtype: dict
        """

        # initialize results
        result_dict = {"result": None, "error": None, "prompt": None, "response_json": None, "response": None}

        # check to see if we meet the criteria for this follow-up
        if condition_func(response_dict, condition_key, condition_value):
            # if so, format the question
            followup_question = prompt_template.format(**response_dict)
            prompt = followup_question.strip()
            result_dict["prompt"] = prompt

            # ask question to the LLM, retrying the appropriate number of times
            attempt = 0
            while True:
                attempt += 1
                try:
                    # ask our question and record the result
                    result = await llm_chain.ainvoke({"question": followup_question})
                    if chat_history is not None:
                        chat_history.append((followup_question, result["answer"]))
                    break
                except Exception as e:
                    self.logger.error(f"Error occurred asking follow-up (attempt {attempt}): {str(e)}")
                    if attempt < self.max_retries:
                        # wait before retrying
                        await asyncio.sleep(5)
                    else:
                        # maximum attempts reached, return error
                        result_dict["result"] = "error"
                        result_dict["error"] = f"Max retries reached on question. Last error: {str(e)}"
                        return result_dict

            # parse result as JSON
            json_response = result["answer"].strip()
            result_dict["response_json"] = json_response
            if json_response != "{}":
                try:
                    result_dict["response"] = json.loads(EvaluationEngine.trim_json(json_response))
                    result_dict["result"] = "success"
                except Exception as e:
                    error_message = f"Error occurred parsing LLM's JSON response ({str(e)}); raw JSON: {json_response}"
                    self.logger.error(error_message)
                    result_dict["error"] = error_message
                    result_dict["result"] = "error"
        else:
            result_dict["result"] = "skipped"

        # return assembled result
        return result_dict


class EvaluationLens:
    """Class for instrument evaluation lens, which is used to conduct a particular type of evaluation."""

    evaluation_engine: EvaluationEngine
    task_system_prompt_template: str
    question_template: str
    followups: list[dict]
    evaluation_result: dict

    def __init__(self, task_system_prompt_template: str, question_template: str, followups: list[dict],
                 evaluation_engine: EvaluationEngine):
        """
        Initialize evaluation lens.

        :param task_system_prompt_template: System prompt template to use for the evaluation chain. This can include the
            {survey_context} and {survey_locations} variables to include information about the survey context. It
            should specify a specific JSON format for all responses.
        :type task_system_prompt_template: str
        :param question_template: Initial question to ask, to begin the evaluation chain. This should include the
            {survey_excerpt} variable to include the appropriate excerpt being evaluated.
        :type question_template: str
        :param followups: List of follow-up questions to ask, based on the JSON response to the initial question.
            Should be a list of dicts, each with a value for each of the following keys:
            "condition_func": function that returns True when the follow-up should be asked and False when it
            shouldn't (should take the following parameters: response_dict: dict, condition_key: str,
            condition_value: value to check against the one in the response dict);
            "condition_key": the key in the response dict to check;
            "condition_value": the value to check against the one in the response dict;
            "prompt_template": the template for the follow-up question to ask (which can include variables from
            the response dict).
        :type followups: list[dict]
        :param evaluation_engine: Evaluation engine instance to use for conducting evaluation.
        :type evaluation_engine: EvaluationEngine
        """

        # initialize object from constructor parameters
        self.task_system_prompt_template = task_system_prompt_template
        if evaluation_engine.extra_evaluation_instructions:
            self.task_system_prompt_template += (f"\n\nAlso follow these additional instructions:\n\n"
                                                 f"{evaluation_engine.extra_evaluation_instructions}")
        self.question_template = question_template
        self.followups = followups
        self.evaluation_engine = evaluation_engine
        self.evaluation_result = {}

    def evaluate(self, chat_history=None, **kwargs) -> dict:
        """
        Run an evaluation chain (synchronously).

        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :param kwargs: Keyword arguments to use for formatting the task system prompt and question.
        :type kwargs: Any
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        # call evaluation engine to run evaluation chain
        self.evaluation_result = self.evaluation_engine.run_evaluation_chain(
            task_system_prompt=self.task_system_prompt_template.format(**kwargs),
            question=self.question_template.format(**kwargs),
            followups=self.followups, chat_history=chat_history)
        return self.evaluation_result
    
    async def a_evaluate(self, chat_history=None, **kwargs) -> dict:
        """
        Run an evaluation chain (asynchronously).

        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :param kwargs: Keyword arguments to use for formatting the task system prompt and question.
        :type kwargs: Any
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        # call evaluation engine to run evaluation chain
        self.evaluation_result = await self.evaluation_engine.a_run_evaluation_chain(
            task_system_prompt=self.task_system_prompt_template.format(**kwargs),
            question=self.question_template.format(**kwargs),
            followups=self.followups, chat_history=chat_history)
        return self.evaluation_result

    def format_result(self, result: dict | None = None, minimum_importance: int = 0) -> str:
        """
        Format the evaluation result as a human-readable string.

        :param result: Evaluation result to format (or None to use the evaluation_result attribute).
        :type result: dict | None
        :param minimum_importance: Minimum importance score for filtering results (defaults to 0, which doesn't filter).
        :type minimum_importance: int
        :return: Formatted evaluation result.
        :rtype: str
        """

        # use evaluation_result attribute if no result is passed
        if result is not None:
            result_to_format = result
        else:
            result_to_format = self.evaluation_result[0]

        # raise error if asked to filter for importance
        if minimum_importance > 0:
            raise NotImplementedError("No base class implementation for filtering by importance.")

        # return empty string if no result is available to format
        if result_to_format is None:
            return ""

        # format result as a list of key-value pairs
        formatted_results = []
        for key, value in result_to_format.items():
            if isinstance(value, list):
                value = [str(item) for item in value]
                if len(value) == 0:
                    value_str = ""
                elif len(value) == 1:
                    value_str = str(value[0])
                else:
                    value_str = ', '.join(value)
                formatted_results.append(f"{key}: {value_str}")
            else:
                formatted_results.append(f"{key}: {value}")

        # return the formatted result with each key-value pair on its own line
        return '\n'.join(formatted_results)

    def standardize_result(self, result: dict | None = None) -> list[dict]:
        """
        Reorganize the evaluation result into a list of recommendations in a standardized format.

        :param result: Evaluation result to format (or None to use the evaluation_result attribute).
        :type result: dict | None
        :return: List of recommendations, each of which is a dict with the following keys: importance (int 1-5),
            replacement_original (str), replacement_suggested (str), explanation (str).
        :rtype: list[dict]
        """

        raise NotImplementedError("No base class implementation for standardize_result().")

    @staticmethod
    def condition_is_value(response_dict: dict, condition_key: str, condition_value: str) -> bool:
        """
        Check if a condition is met in a given response dictionary: "is value" (string match).

        :param response_dict: Response dictionary to check.
        :type response_dict: dict
        :param condition_key: Key to check in response dictionary.
        :type condition_key: str
        :param condition_value: Value to check for in response dictionary.
        :type condition_value: str
        :return: True if condition is met, False otherwise.
        :rtype: bool
        """

        return condition_key in response_dict and response_dict[condition_key] == condition_value

    @staticmethod
    def condition_is_not_value(response_dict: dict, condition_key: str, condition_value: str) -> bool:
        """
        Check if a condition is met in a given response dictionary: "is not value" (string doesn't match).

        :param response_dict: Response dictionary to check.
        :type response_dict: dict
        :param condition_key: Key to check in response dictionary.
        :type condition_key: str
        :param condition_value: Value to check for in response dictionary.
        :type condition_value: str
        :return: True if condition is met, False otherwise.
        :rtype: bool
        """

        return condition_key in response_dict and response_dict[condition_key] != condition_value

    @staticmethod
    def condition_is_in_list(response_dict: dict, condition_key: str, condition_value: str) -> bool:
        """
        Check if a condition is met in a given response dictionary: "is in list" (string in list of strings).

        :param response_dict: Response dictionary to check.
        :type response_dict: dict
        :param condition_key: Key to check in response dictionary.
        :type condition_key: str
        :param condition_value: Value to check for in response dictionary.
        :type condition_value: str
        :return: True if condition is met, False otherwise.
        :rtype: bool
        """

        return condition_key in response_dict and response_dict[condition_key] in condition_value

    @staticmethod
    def condition_is_not_in_list(response_dict: dict, condition_key: str, condition_value: str) -> bool:
        """
        Check if a condition is met in a given response dictionary: "is not in list" (string not in list of strings).

        :param response_dict: Response dictionary to check.
        :type response_dict: dict
        :param condition_key: Key to check in response dictionary.
        :type condition_key: str
        :param condition_value: Value to check for in response dictionary.
        :type condition_value: str
        :return: True if condition is met, False otherwise.
        :rtype: bool
        """

        return condition_key in response_dict and response_dict[condition_key] not in condition_value

    @staticmethod
    def condition_list_has_greater_than_elements(response_dict: dict, condition_key: str, condition_value: int) -> bool:
        """
        Check if a condition is met in a given response dictionary: "list has greater than n elements".

        :param response_dict: Response dictionary to check.
        :type response_dict: dict
        :param condition_key: Key to check in response dictionary.
        :type condition_key: str
        :param condition_value: Value to check for in response dictionary.
        :type condition_value: int
        :return: True if condition is met, False otherwise.
        :rtype: bool
        """

        if condition_key in response_dict and isinstance(response_dict[condition_key], list):
            return len(response_dict[condition_key]) > condition_value
        return False

    @staticmethod
    def condition_list_has_greater_or_equal_elements(response_dict: dict, condition_key: str, condition_value: int) \
            -> bool:
        """
        Check if a condition is met in a given response dictionary: "list has greater than or equal to n elements".

        :param response_dict: Response dictionary to check.
        :type response_dict: dict
        :param condition_key: Key to check in response dictionary.
        :type condition_key: str
        :param condition_value: Value to check for in response dictionary.
        :type condition_value: int
        :return: True if condition is met, False otherwise.
        :rtype: bool
        """

        if condition_key in response_dict and isinstance(response_dict[condition_key], list):
            return len(response_dict[condition_key]) >= condition_value
        return False

    @staticmethod
    def condition_list_has_less_than_elements(response_dict: dict, condition_key: str, condition_value: int) \
            -> bool:
        """
        Check if a condition is met in a given response dictionary: "list has less than n elements".

        :param response_dict: Response dictionary to check.
        :type response_dict: dict
        :param condition_key: Key to check in response dictionary.
        :type condition_key: str
        :param condition_value: Value to check for in response dictionary.
        :type condition_value: int
        :return: True if condition is met, False otherwise.
        :rtype: bool
        """

        if condition_key in response_dict and isinstance(response_dict[condition_key], list):
            return len(response_dict[condition_key]) < condition_value
        return False

    @staticmethod
    def condition_list_has_less_or_equal_elements(response_dict: dict, condition_key: str, condition_value: int) \
            -> bool:
        """
        Check if a condition is met in a given response dictionary: "list has less than or equal to n elements".

        :param response_dict: Response dictionary to check.
        :type response_dict: dict
        :param condition_key: Key to check in response dictionary.
        :type condition_key: str
        :param condition_value: Value to check for in response dictionary.
        :type condition_value: int
        :return: True if condition is met, False otherwise.
        :rtype: bool
        """

        if condition_key in response_dict and isinstance(response_dict[condition_key], list):
            return len(response_dict[condition_key]) <= condition_value
        return False



# Adapted from the Kor version at https://github.com/eyurtsev/kor
# Kor's license:
#
# MIT License
#
# Copyright (c) 2023 Eugene Yurtsev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Load and chunk HTMLs with potential pre-processing to clean the html."""

import re
from typing import Tuple

from langchain.schema import Document

# Regular expression pattern to detect multiple new lines in a row with optional
# whitespace in between
CONSECUTIVE_NEW_LINES = re.compile(r"\n(\s*\n)+", flags=re.UNICODE)


def _get_mini_html(html: str, *, tags_to_remove: Tuple[str, ...] = tuple()) -> str:
    """Clean up HTML tags."""
    try:
        from bs4 import BeautifulSoup
    except ImportError:
        raise ImportError(
            "Please install BeautifulSoup to use the HTML document processor. "
            "You can do so by running `pip install beautifulsoup4`."
        )
    # Parse the HTML document using BeautifulSoup
    soup = BeautifulSoup(html, "html.parser")

    # Remove all CSS stylesheets
    for stylesheet in soup.find_all("link", rel="stylesheet"):
        stylesheet.extract()

    for tag_to_remove in tags_to_remove:
        # Remove all matching tags
        for tag in soup.find_all(tag_to_remove):
            tag.extract()

    new_html = repr(soup)
    return new_html


def _clean_html(html: str, *, tags_to_remove: Tuple[str, ...] = tuple()) -> str:
    """Clean up HTML and convert to markdown using markdownify."""
    try:
        import markdownify
    except ImportError:
        raise ImportError(
            "Please install markdownify to use the HTML document processor. "
            "You can do so by running `pip install markdownify`."
        )

    html = _get_mini_html(html, tags_to_remove=tags_to_remove)
    md = markdownify.markdownify(html)
    return CONSECUTIVE_NEW_LINES.sub("\n\n", md).strip()


class MarkdownifyHTMLProcessor(object):
    """A preprocessor to clean HTML and convert to markdown using markdownify."""

    def __init__(
        self,
        tags_to_remove: Tuple[str, ...] = ("svg", "img", "script", "style"),
    ) -> None:
        """Initialize the preprocessor.

        Args:
            tags_to_remove: A tuple of tags to remove from the HTML
        """
        self.tags_to_remove = tags_to_remove

    def process(self, document: Document) -> Document:
        """Clean up HTML and convert to markdown using markdownify.

        Args:
            document: a document with HTML content

        Returns:
            The cleaned HTML
        """
        new_document = document.copy()
        new_document.page_content = _clean_html(
            document.page_content, tags_to_remove=self.tags_to_remove
        )
        return new_document



#  Copyright (c) 2024 Higher Bar AI, PBC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .evaluation_engine import EvaluationEngine, EvaluationLens
from .core_evaluation_lenses import (BiasEvaluationLens, PhrasingEvaluationLens, TranslationEvaluationLens,
                                     ValidatedInstrumentEvaluationLens)

# report our current version, as installed
from importlib.metadata import version
try:
    __version__ = version("surveyeval")
except Exception:
    # (ignore exceptions when developing and testing locally)
    pass



#  Copyright (c) 2023-24 Higher Bar AI, PBC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Core set of instrument evaluation lenses."""

from surveyeval import EvaluationEngine, EvaluationLens
from overrides import overrides
import json


class PhrasingEvaluationLens(EvaluationLens):
    """
    Lens for identifying phrasing issues that might be flagged during piloting or cognitive interviewing.
    """

    def __init__(self, evaluation_engine: EvaluationEngine):
        """
        Override default constructor to provide lens-specific prompt template and followup questions.

        :param evaluation_engine: EvaluationEngine instance to use for evaluation.
        :type evaluation_engine: EvaluationEngine
        """

        lens_system_prompt_template = """You are an AI designed to evaluate questionnaires and other survey instruments
used by researchers and M&E professionals. You are an expert in survey methodology with training equivalent to a member
of the American Association for Public Opinion Research (AAPOR) with a Ph.D. in survey methodology from University of
Michigan’s Institute for Social Research. You consider primarily the content, context, and questions provided to you,
and then content and methods from the most widely-cited academic publications and public and nonprofit research
organizations.

You always give truthful, factual answers. When asked to give your response in a specific format, you always give your
answer in the exact format requested. You never give offensive responses. If you don’t know the answer to a question,
you truthfully say you don’t know.

You will be given an excerpt from a questionnaire or survey instrument between |!| and |!| delimiters. You will also
be given a specific question from that excerpt to evaluate between |@| and |@| delimiters. Evaluate the question only,
but also consider its context within the larger excerpt.

The broader context and location(s) for that excerpt are as follows. Consider these when evaluating the question.

Survey context: {survey_context}

Survey locations: {survey_locations}

Assume that this survey will be administered by a trained enumerator who asks each question and reads each prompt or
instruction as indicated in the excerpt. Your job is to anticipate the phrasing or translation issues that would be
identified in a rigorous process of pre-testing (with cognitive interviewing) and piloting.

When evaluating the question, DO: (1) ensure that the question will be understandable by substantially all 
respondents, based on the survey context and locations above; (2) consider the language, location, and survey context 
(with the appropriate location for the language coming from the "Survey locations" list above); (3) consider the 
question in the context of the excerpt, including any instructions, related questions, or prompts that precede it; (4) 
ignore question numbers and formatting; and (5) assume that code to refer to earlier responses like [QUESTION] or 
${{{{question}}}} is okay as it is.

When evaluating the question, DON'T: recommend translating something into another language (i.e., suggestions for
rephrasing should always be in the same language as the original text).

Only recommend changes in the overall structure of a question (e.g., changing from multiple choice to open-ended or 
splitting one question into multiple) when it will substantially improve the quality of the data collected.

Respond in JSON format with all of the following fields:

* Phrases: a list containing all phrases from the excerpt that pre-testing or piloting is likely to identify
as problematic (each phrase should be an exact quote)

* Number of phrases: the exact number of phrases in Phrases

* Recommendations: a list containing suggested replacement phrases, one for each of the phrases in Phrases (in the same
order as Phrases; each replacement phrase should be an exact quote that can exactly replace the corresponding phrase in
Phrases; and each replacement phrase should be in the same language as the original phrase)

* Explanations: a list containing explanations for why the authors should consider revising each phrase, one for each
of the phrases in Phrases (in the same order as Phrases). Do not repeat the entire phrase in the explanation, but feel
free to reference specific words or parts as needed.

* Severities: a list containing the severity of each identified issue, one for each of the phrases in Phrases (in the
same order as Phrases); each severity should be expressed as a number on a scale from 1 for the least severe issues
(minor phrasing issues that are very unlikely to substantively affect responses) to 5 for the most severe issues
(problems that are likely to substantively affect responses in a way that introduces bias and/or variance)"""

        lens_question_template = """Excerpt: |!|{survey_excerpt}|!|\n\nQuestion: |@|{survey_question}|@|"""

        lens_followups = [
            {
                'condition_func': EvaluationLens.condition_list_has_less_than_elements,
                'condition_key': 'Phrases',
                'condition_value': 1,
                'prompt_template': """Are you certain that there are no phrasing issues likely to be identified in
cognitive interviewing or piloting? If appropriate, please respond with a revised JSON response (including all fields).
If you have no changes to propose, respond with an empty JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_list_has_greater_or_equal_elements,
                'condition_key': 'Phrases',
                'condition_value': 1,
                'prompt_template': """Are you certain that (1) the Phrases, Recommendations, Explanations, and
Severities lists each have exactly {Number of phrases} elements, in the same parallel order; (2) every element of the
Severities list is a 1, 2, 3, 4, or 5, depending on the severity of the identified issue (with the most minor phrasing
issues receiving a 1 and the most serious phrasing issues receiving a 5); (3) you didn't try to replace code for
prior responses like [QUESTION] or ${{{{question}}}}; (4) your replacement phrase is in the same language as the
original phrase; and (5) your recommendation concerns the specific question you were asked to evaluate? If appropriate, 
please respond with a revised JSON response (including all fields). If you have no changes to propose, respond with an 
empty JSON response of {{}}."""
            }
        ]

        # call super constructor
        super().__init__(lens_system_prompt_template, lens_question_template, lens_followups, evaluation_engine)

    @overrides
    def evaluate(self, chat_history: list = None, survey_context: str = "", survey_locations: str = "",
                 survey_excerpt: str = "", survey_question: str = "", **kwargs) \
            -> dict:
        """
        Override default evaluate method.

        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :param survey_context: Information about the survey context.
        :type survey_context: str
        :param survey_locations: Information about the survey location(s).
        :type survey_locations: str
        :param survey_excerpt: Excerpt from the survey instrument (for context).
        :type survey_excerpt: str
        :param survey_question: Specific question to focus on.
        :type survey_question: str
        :param kwargs: Keyword arguments to use for formatting the task system prompt and question.
        :type kwargs: Any
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        return super().evaluate(chat_history=chat_history, survey_context=survey_context,
                                survey_locations=survey_locations,
                                survey_excerpt=EvaluationEngine.clean_whitespace(survey_excerpt),
                                survey_question=EvaluationEngine.clean_whitespace(survey_question), **kwargs)

    @overrides
    async def a_evaluate(self, chat_history: list = None, survey_context: str = "", survey_locations: str = "",
                         survey_excerpt: str = "", survey_question: str = "", **kwargs) \
            -> dict:
        """
        Override default a_evaluate method.

        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :param survey_context: Information about the survey context.
        :type survey_context: str
        :param survey_locations: Information about the survey location(s).
        :type survey_locations: str
        :param survey_excerpt: Excerpt from the survey instrument to evaluate.
        :type survey_excerpt: str
        :param survey_question: Specific question to focus on.
        :type survey_question: str
        :param kwargs: Keyword arguments to use for formatting the task system prompt and question.
        :type kwargs: Any
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        return await super().a_evaluate(chat_history=chat_history, survey_context=survey_context,
                                        survey_locations=survey_locations,
                                        survey_excerpt=EvaluationEngine.clean_whitespace(survey_excerpt),
                                        survey_question=EvaluationEngine.clean_whitespace(survey_question),
                                        **kwargs)

    def format_result(self, result: dict | None = None, minimum_importance: int = 0) -> str:
        """
        Format the evaluation result as a human-readable string.

        :param result: Evaluation result to format (or None to use the evaluation_result attribute).
        :type result: dict | None
        :param minimum_importance: Minimum importance score for filtering results (defaults to 0, which doesn't filter).
        :type minimum_importance: int
        :return: Formatted evaluation result.
        :rtype: str
        """

        # use evaluation_result attribute if no result is passed
        if result is not None:
            result_to_format = result
        else:
            result_to_format = self.evaluation_result

        # return empty string if no result is available to format
        if result_to_format is None:
            return ""

        # format and return result
        try:
            formatted_result = ""
            sorted_indices = sorted(range(len(result_to_format["Severities"])),
                                    key=lambda idx: result_to_format["Severities"][idx], reverse=True)
            for i in sorted_indices:
                severity = result_to_format['Severities'][i]
                # only report findings with severity greater than or equal to minimum_importance
                if severity >= minimum_importance:
                    formatted_result += f"Severity {severity} finding (out of 5):\n\n"
                    formatted_result += f"{result_to_format['Explanations'][i]}\n\n"
                    formatted_result += f"Recommend replacing: {result_to_format['Phrases'][i]}\n"
                    formatted_result += f"          With this: {result_to_format['Recommendations'][i]}\n\n"
        except Exception as e:
            # include exception in returned results, with raw JSON results
            formatted_result = (f"Error occurred formatting result: {str(e)}\n\n"
                                f"Raw JSON: {json.dumps(result_to_format, indent=4)}")
            self.evaluation_engine.logger.error(formatted_result)
        return formatted_result

    def standardize_result(self, result: dict | None = None) -> list[dict]:
        """
        Reorganize the evaluation result into a list of recommendations in a standardized format.

        :param result: Evaluation result to format (or None to use the evaluation_result attribute).
        :type result: dict | None
        :return: List of recommendations, each of which is a dict with the following keys: importance (int 1-5),
            replacement_original (str), replacement_suggested (str), explanation (str).
        :rtype: list[dict]
        """

        # use evaluation_result attribute if no result is passed
        if result is not None:
            result_to_format = result
        else:
            result_to_format = self.evaluation_result

        # organize and return result
        return_list = []
        try:
            # first make sure we have a result to organize
            if result_to_format is not None and result_to_format["Number of phrases"] > 0:
                # sort the results by severity
                sorted_indices = sorted(range(len(result_to_format["Severities"])),
                                        key=lambda idx: result_to_format["Severities"][idx], reverse=True)
                # run through each result, appending to our return list
                for i in sorted_indices:
                    return_list.append({
                        'importance': result_to_format['Severities'][i],
                        'replacement_original': result_to_format['Phrases'][i],
                        'replacement_suggested': result_to_format['Recommendations'][i],
                        'explanation': result_to_format['Explanations'][i]
                    })
        except Exception as e:
            raise RuntimeError(f"Error occurred formatting result: {str(e)}\n\n"
                               f"Raw JSON: {json.dumps(result_to_format, indent=4)}") from e
        return return_list


class ValidatedInstrumentEvaluationLens(EvaluationLens):
    """
    Lens for identifying validated questions, instruments, or tools that either were used or could be used to measure
    what the excerpt is attempting to measure.
    """

    def __init__(self, evaluation_engine: EvaluationEngine):
        """
        Override default constructor to provide lens-specific prompt template and followup questions.

        :param evaluation_engine: EvaluationEngine instance to use for evaluation.
        :type evaluation_engine: EvaluationEngine
        """

        lens_system_prompt_template = """You are an AI designed to evaluate questionnaires and other survey instruments
used by researchers and M&E professionals. You are an expert in survey methodology with training equivalent to a member
of the American Association for Public Opinion Research (AAPOR) with a Ph.D. in survey methodology from University of
Michigan’s Institute for Social Research. You consider primarily the content, context, and questions provided to you,
and then content and methods from the most widely-cited academic publications and public and nonprofit research
organizations.

You always give truthful, factual answers. When asked to give your response in a specific format, you always give your
answer in the exact format requested. You never give offensive responses. If you don’t know the answer to a question,
you truthfully say you don’t know.

You will be given an excerpt from a questionnaire or survey instrument between |@| and |@| delimiters. The context
and location(s) for that excerpt are as follows:

Survey context: {survey_context}

Survey locations: {survey_locations}

Your job is to respond with your evaluation in JSON format with all of the following fields:

* Measuring: a list containing one or more short strings describing what the excerpt seems to be attempting to measure

* Replication: 1 if the excerpt includes an exact replication of a validated question, instrument, or tool commonly
used for measuring what the excerpt is attempting to measure; otherwise 0. Only answer 1 if the excerpt includes all of
the same questions and response options as the validated version.

* Replication name: if Replication is 1, the name of the validated question, instrument, or tool that has been
replicated; otherwise an empty string

* Replication URL: if Replication is 1, a URL to learn more about the validated question, instrument, or tool that has
been replicated; otherwise an empty string. Only give a URL if you are confident that the URL is the right place to go
to learn more, otherwise describe where to go to learn more.

* Replication explanation: if Replication is 1, a short description of how the excerpt includes the validated version;
otherwise an empty string

* Adaptation: 1 if the excerpt includes an adapted version of a validated question, instrument, or tool commonly used
for measuring what the excerpt is attempting to measure; otherwise 0. Only answer 1 if the excerpt includes strong
similarity to the validated version, but not exactly the same questions or response options. If the excerpt includes
only weak or superficial similarity to the validated version, answer 0.

* Adaptation name: if Adaptation is 1, the name of the validated question, instrument, or tool that has been adapted;
otherwise an empty string

* Adaptation URL: if Adaptation is 1, a URL to learn more about the validated question, instrument, or tool that has
been adapted; otherwise an empty string. Only give a URL if you are confident that the URL is the right place to go to
learn more, otherwise describe where to go to learn more.

* Adaptation explanation: if Adaptation is 1, a short description of how the excerpt is similar to and different from
the validated version; otherwise an empty string

* Recommendation: 1 if you would recommend that the author consider using a validated question, instrument, or tool
commonly used for measuring what the excerpt is attempting to measure (or a different one, if Replication or Adaptation
is 1), rather than the current approach; otherwise 0

* Recommendation name: if Recommendation is 1, the name of the validated question, instrument, or tool you would
recommend; otherwise an empty string

* Recommendation URL: if Recommendation is 1, a URL to learn more about the validated question, instrument, or tool you
would recommend; otherwise an empty string. Only give a URL if you are confident that the URL is the right place to go
to learn more, otherwise describe where to go to learn more.

* Recommendation explanation: if Recommendation is 1, a short description of why the author should consider the
validated version proposed; otherwise an empty string

* Recommendation strength: if Recommendation is 1, a number from 1 to 5 to indicate the strength of the recommendation,
with 1 being the weakest possible recommendation and 5 being the strongest possible recommendation; otherwise an empty
string. In deciding on a Recommendation strength, consider both what is being recommended and how good a fit it might
be to the survey context and locations."""

        lens_question_template = """Excerpt: |@|{survey_excerpt}|@|"""

        lens_followups = [
            {
                'condition_func': EvaluationLens.condition_is_value,
                'condition_key': 'Replication',
                'condition_value': 1,
                'prompt_template': """Are you certain that the excerpt contains an exact replication of
{Replication name}, including the same question(s) and answer option(s)? If the version in the excerpt is different in
some way, then Replication should be 0 and Adaptation should be 1. If appropriate, please respond with a revised JSON
response (including all fields). If you have no changes to propose, respond with an empty JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_is_value,
                'condition_key': 'Adaptation',
                'condition_value': 1,
                'prompt_template': """Are you certain that the excerpt contains an adaptation of {Adaptation name},
meaning that it exhibits strong similarity to the validated version, but not exactly the same questions or response
options? If the excerpt includes only weak or superficial similarity to the validated version, it should not be
considered an adaptation. And if its questions and response options are exactly the same as a validated version, then
it should be considered a Replication instead. If appropriate, please respond with a revised JSON response (including
all fields). If you have no changes to propose, respond with an empty JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_is_not_value,
                'condition_key': 'Replication URL',
                'condition_value': "",
                'prompt_template': """Are you certain that the Replication URL you supplied is the best place for the
author to learn more? If not, please respond with a revised JSON response with a Replication URL that includes a
different URL or a short description of where the author should go to learn more (including all fields). If you have no
changes to propose, respond with an empty JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_is_not_value,
                'condition_key': 'Adaptation URL',
                'condition_value': "",
                'prompt_template': """Are you certain that the Adaptation URL you supplied is the best place for the
author to learn more? If not, please respond with a revised JSON response with a Adaptation URL that includes a
different URL or a short description of where the author should go to learn more (including all fields). If you have no
changes to propose, respond with an empty JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_is_not_value,
                'condition_key': 'Recommendation URL',
                'condition_value': "",
                'prompt_template': """Are you certain that the Recommendation URL you supplied is the best place for
the author to learn more? If not, please respond with a revised JSON response with a Recommendation URL that includes a
different URL or a short description of where the author should go to learn more (including all fields). If you have no
changes to propose, respond with an empty JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_is_value,
                'condition_key': 'Recommendation',
                'condition_value': 1,
                'prompt_template': """Are you certain that the Recommendation explanation and Recommendation strength
you supplied is correct, and that the two are consistent with one another? If not, please respond with a revised JSON
response that revises the recommendation details as appropriate (including all fields). If you have no changes to
propose, respond with an empty JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_is_not_in_list,
                'condition_key': 'Recommendation strength',
                'condition_value': [1, 2, 3, 4, 5],
                'prompt_template': """The Recommendation strength should be a number from 1 to 5 to indicate the
strength of the recommendation, with 1 being the weakest possible recommendation and 5 being the strongest possible
recommendation. Please respond with a revised JSON response that revises the Recommendation strength to be a number
from 1 to 5, based on the strength of the recommendation (including all fields)."""
            }
        ]

        # call super constructor
        super().__init__(lens_system_prompt_template, lens_question_template, lens_followups, evaluation_engine)

    @overrides
    def evaluate(self, chat_history: list = None, survey_context: str = "", survey_locations: str = "",
                 survey_excerpt: str = "", **kwargs) -> dict:
        """
        Override default evaluate method.

        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :param survey_context: Information about the survey context.
        :type survey_context: str
        :param survey_locations: Information about the survey location(s).
        :type survey_locations: str
        :param survey_excerpt: Excerpt from the survey instrument to evaluate.
        :type survey_excerpt: str
        :param kwargs: Keyword arguments to use for formatting the task system prompt and question.
        :type kwargs: Any
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        return super().evaluate(chat_history=chat_history, survey_context=survey_context,
                                survey_locations=survey_locations,
                                survey_excerpt=EvaluationEngine.clean_whitespace(survey_excerpt), **kwargs)

    @overrides
    async def a_evaluate(self, chat_history: list = None, survey_context: str = "", survey_locations: str = "",
                         survey_excerpt: str = "", **kwargs) -> dict:
        """
        Override default a_evaluate method.

        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :param survey_context: Information about the survey context.
        :type survey_context: str
        :param survey_locations: Information about the survey location(s).
        :type survey_locations: str
        :param survey_excerpt: Excerpt from the survey instrument to evaluate.
        :type survey_excerpt: str
        :param kwargs: Keyword arguments to use for formatting the task system prompt and question.
        :type kwargs: Any
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        return await super().a_evaluate(chat_history=chat_history, survey_context=survey_context,
                                        survey_locations=survey_locations,
                                        survey_excerpt=EvaluationEngine.clean_whitespace(survey_excerpt), **kwargs)

    def format_result(self, result: dict | None = None, minimum_importance: int = 0) -> str:
        """
        Format the evaluation result as a human-readable string.

        :param result: Evaluation result to format (or None to use the evaluation_result attribute).
        :type result: dict | None
        :param minimum_importance: Minimum importance score for filtering results (defaults to 0, which doesn't filter).
        :type minimum_importance: int
        :return: Formatted evaluation result.
        :rtype: str
        """

        # use evaluation_result attribute if no result is passed
        if result is not None:
            result_to_format = result
        else:
            result_to_format = self.evaluation_result

        # return empty string if no result is available to format
        if result_to_format is None:
            return ""

        # format and return result
        try:
            if minimum_importance > 0 and (not result_to_format["Recommendation"]
                                           or result_to_format["Recommendation strength"] < minimum_importance):
                # filter this result since as not important enough
                return ""

            if len(result_to_format["Measuring"]) == 0:
                measuring_str = ""
            elif len(result_to_format["Measuring"]) == 1:
                measuring_str = str(result_to_format["Measuring"][0])
            else:
                measuring_str = ', '.join(result_to_format["Measuring"])
            formatted_result = f"FYI: Excerpt likely attempting to measure {measuring_str}\n\n"

            if result_to_format["Replication"]:
                formatted_result = (f'FYI: Excerpt likely replication of "{result_to_format["Replication name"]}". '
                                    f'{result_to_format["Replication explanation"]} (Learn more: '
                                    f'{result_to_format["Replication URL"]})\n\n')
            if result_to_format["Adaptation"]:
                formatted_result = (f'FYI: Excerpt likely adaptation of "{result_to_format["Adaptation name"]}". '
                                    f'{result_to_format["Adaptation explanation"]} (Learn more: '
                                    f'{result_to_format["Adaptation URL"]})\n\n')

            if result_to_format["Recommendation"]:
                formatted_result = (f'Recommendation (strength {result_to_format["Recommendation strength"]} '
                                    f'out of 5):\n\n'
                                    f'Consider adapting: {result_to_format["Recommendation name"]}\n'
                                    f'Explanation: {result_to_format["Recommendation explanation"]}\n'
                                    f'Learn more: {result_to_format["Recommendation URL"]}\n\n')
        except Exception as e:
            # include exception in returned results, with raw JSON results
            formatted_result = (f"Error occurred formatting result: {str(e)}\n\n"
                                f"Raw JSON: {json.dumps(result_to_format, indent=4)}")
            self.evaluation_engine.logger.error(formatted_result)
        return formatted_result

    def standardize_result(self, result: dict | None = None) -> list[dict]:
        """
        Reorganize the evaluation result into a list of recommendations in a standardized format.

        :param result: Evaluation result to format (or None to use the evaluation_result attribute).
        :type result: dict | None
        :return: List of recommendations, each of which is a dict with the following keys: importance (int 1-5),
            replacement_original (str), replacement_suggested (str), explanation (str).
        :rtype: list[dict]
        """

        # use evaluation_result attribute if no result is passed
        if result is not None:
            result_to_format = result
        else:
            result_to_format = self.evaluation_result

        # organize and return result
        return_list = []
        try:
            # first make sure we have a meaningful result to organize
            if result_to_format is not None and result_to_format["Recommendation"]:
                # start with the recommendation itself
                explanation = (f"Consider adapting {result_to_format['Recommendation name']}. "
                               f"{result_to_format['Recommendation explanation']} "
                               f"(Learn more: {result_to_format['Recommendation URL']})")

                # add information about what the excerpt is attempting to measure (if any)
                if len(result_to_format["Measuring"]) == 0:
                    measuring_str = ""
                elif len(result_to_format["Measuring"]) == 1:
                    measuring_str = str(result_to_format["Measuring"][0])
                else:
                    measuring_str = ', '.join(result_to_format["Measuring"])
                if measuring_str:
                    explanation += f"\n\nFYI, you appear to be measuring {measuring_str}."

                if result_to_format["Replication"]:
                    explanation += (f'\n\nFYI, this could be a replication of '
                                    f'"{result_to_format["Replication name"]}". '
                                    f'{result_to_format["Replication explanation"]} (Learn more: '
                                    f'{result_to_format["Replication URL"]})')
                if result_to_format["Adaptation"]:
                    explanation += (f'\n\nFYI, this could be an adaptation of '
                                    f'"{result_to_format["Adaptation name"]}". '
                                    f'{result_to_format["Adaptation explanation"]} (Learn more: '
                                    f'{result_to_format["Adaptation URL"]})')

                # append our one and only recommendation to the return list
                return_list.append({
                    'importance': result_to_format['Recommendation strength'],
                    'replacement_original': "",
                    'replacement_suggested': "",
                    'explanation': explanation
                })
        except Exception as e:
            raise RuntimeError(f"Error occurred formatting result: {str(e)}\n\n"
                               f"Raw JSON: {json.dumps(result_to_format, indent=4)}") from e
        return return_list


class TranslationEvaluationLens(EvaluationLens):
    """
    Lens for identifying translation issues that could lead to differing response patterns from respondents.
    """

    def __init__(self, evaluation_engine: EvaluationEngine):
        """
        Override default constructor to provide lens-specific prompt template and followup questions.

        :param evaluation_engine: EvaluationEngine instance to use for evaluation.
        :type evaluation_engine: EvaluationEngine
        """

        lens_system_prompt_template = """You are an AI designed to evaluate questionnaires and other survey instruments
used by researchers and M&E professionals. You are an expert in survey methodology with training equivalent to a member
of the American Association for Public Opinion Research (AAPOR) with a Ph.D. in survey methodology from University of
Michigan’s Institute for Social Research. You consider primarily the content, context, and questions provided to you,
and then content and methods from the most widely-cited academic publications and public and nonprofit research
organizations.

You always give truthful, factual answers. When asked to give your response in a specific format, you always give your
answer in the exact format requested. You never give offensive responses. If you don’t know the answer to a question,
you truthfully say you don’t know.

You will be given an excerpt from a questionnaire or survey instrument between |@| and |@| delimiters. The context and
location(s) for that excerpt are as follows:

Survey context: {survey_context}

Survey locations: {survey_locations}

The excerpt will include the same questions and response options in two languages, one labeled as primary and one 
labeled as translated. Assume that this survey will be administered by a trained enumerator who asks each question in a 
single language appropriate to the respondent and reads each prompt or instruction as indicated in the excerpt. Your 
job is to review the excerpt for differences in the translation that could lead to differing response patterns from 
respondents. The goal is for translations to be accurate enough that data collected will be comparable regardless of 
the language of administration.

Assume that each language will be used in the location contexts as indicated in the "Survey locations" details
above, and consider those location contexts (and only those location contexts) in your evaluation. For example, when 
evaluating a question in French or English, consider the appropriate country contexts listed in the "Survey locations", 
as languages might have different dialects and conventions in different countries.

Also assume that translations should be designed for understandability and comparability across locations. Literal
translations might not best achieve the same meaning and response patterns across settings. Ensure that the meaning is
common across languages and locations.

Finally, identify problematic phrases and replacement text in the translated language only. Do not propose any  
changes in the primary language.

Respond in JSON format with all of the following fields:

* Phrases: a list containing all problematic phrases from the translation that you found in your review, where 
the translation does not adequately match the meaning in the primary language (each phrase should be an exact quote 
from the translated language)

* Number of phrases: the exact number of phrases in Phrases

* Recommendations: a list containing suggested replacement phrases, one for each of the phrases in Phrases (in the same
order as Phrases; each replacement phrase should be an exact quote that can exactly replace the corresponding phrase in
Phrases; and each replacement phrase should be in the same translated language as the phrase it replaces)

* Explanations: a list containing explanations for why the phrases are problematic, one for each of the phrases in
Phrases (in the same order as Phrases)

* Severities: a list containing the severity of each identified issue, one for each of the phrases in Phrases (in the
same order as Phrases); each severity should be expressed as a number on a scale from 1 for the least severe issues
(minor phrasing issues that are very unlikely to substantively affect response patterns) to 5 for the most severe
issues (problems that are very likely to substantively affect response patterns in a way that introduces bias and/or
variance)"""

        lens_question_template = """Excerpt: |@|{survey_excerpt}|@|"""

        lens_followups = [
            {
                'condition_func': EvaluationLens.condition_list_has_less_than_elements,
                'condition_key': 'Phrases',
                'condition_value': 1,
                'prompt_template': """Are you sure that there are no inaccurate translations, nor any response options
that are missing or ordered differently in the translated version? If appropriate, please respond with a revised JSON 
response (including all fields). If you have no changes to propose, respond with an empty JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_list_has_greater_or_equal_elements,
                'condition_key': 'Phrases',
                'condition_value': 1,
                'prompt_template': """Differences in the meaning of questions or response options can affect response
patterns in important ways. Are you sure that you didn't miss any cases where differences in meaning could lead to
different response patterns? If appropriate, please respond with a revised JSON response (including all fields). If you
have no changes to propose, respond with an empty JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_list_has_greater_or_equal_elements,
                'condition_key': 'Phrases',
                'condition_value': 1,
                'prompt_template': """If the same response options are not present in all translations in the same
order, response patterns can differ. Are you sure that you didn't miss any cases where response options are missing
from the translation, or in a different order? If appropriate, please respond with a revised JSON response (including 
all fields). If you have no changes to propose, respond with an empty JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_list_has_greater_or_equal_elements,
                'condition_key': 'Phrases',
                'condition_value': 1,
                'prompt_template': """Are you certain that (1) the Phrases, Recommendations, Explanations,
and Severities lists each have exactly {Number of phrases} elements, in the same parallel order; (2) every element
of the Severities list is a 1, 2, 3, 4, or 5, depending on the severity of the issue; and (3) every phrase in Phrases 
and every recommendation in Recommendations is in the translated language? If appropriate, please respond
with a revised JSON response (including all fields). If you have no changes to propose, respond with an empty JSON
response of {{}}."""
            }
        ]

        # call super constructor
        super().__init__(lens_system_prompt_template, lens_question_template, lens_followups, evaluation_engine)

    @overrides
    def evaluate(self, chat_history: list = None, survey_context: str = "", survey_locations: str = "",
                 survey_excerpt: str = "", **kwargs) -> dict:
        """
        Override default evaluate method.

        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :param survey_context: Information about the survey context.
        :type survey_context: str
        :param survey_locations: Information about the survey location(s).
        :type survey_locations: str
        :param survey_excerpt: Excerpt from the survey instrument to evaluate.
        :type survey_excerpt: str
        :param kwargs: Keyword arguments to use for formatting the task system prompt and question.
        :type kwargs: Any
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        return super().evaluate(chat_history=chat_history, survey_context=survey_context,
                                survey_locations=survey_locations,
                                survey_excerpt=EvaluationEngine.clean_whitespace(survey_excerpt), **kwargs)

    @overrides
    async def a_evaluate(self, chat_history: list = None, survey_context: str = "", survey_locations: str = "",
                         survey_excerpt: str = "", **kwargs) -> dict:
        """
        Override default a_evaluate method.

        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :param survey_context: Information about the survey context.
        :type survey_context: str
        :param survey_locations: Information about the survey location(s).
        :type survey_locations: str
        :param survey_excerpt: Excerpt from the survey instrument to evaluate.
        :type survey_excerpt: str
        :param kwargs: Keyword arguments to use for formatting the task system prompt and question.
        :type kwargs: Any
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        return await super().a_evaluate(chat_history=chat_history, survey_context=survey_context,
                                        survey_locations=survey_locations,
                                        survey_excerpt=EvaluationEngine.clean_whitespace(survey_excerpt), **kwargs)

    def format_result(self, result: dict | None = None, minimum_importance: int = 0) -> str:
        """
        Format the evaluation result as a human-readable string.

        :param result: Evaluation result to format (or None to use the evaluation_result attribute).
        :type result: dict | None
        :param minimum_importance: Minimum importance score for filtering results (defaults to 0, which doesn't filter).
        :type minimum_importance: int
        :return: Formatted evaluation result.
        :rtype: str
        """

        # use evaluation_result attribute if no result is passed
        if result is not None:
            result_to_format = result
        else:
            result_to_format = self.evaluation_result

        # return empty string if no result is available to format
        if result_to_format is None:
            return ""

        # format and return result
        try:
            formatted_result = ""
            sorted_indices = sorted(range(len(result_to_format["Severities"])),
                                    key=lambda idx: result_to_format["Severities"][idx], reverse=True)
            for i in sorted_indices:
                severity = result_to_format['Severities'][i]
                # only report findings with severity greater than or equal to minimum_importance
                if severity >= minimum_importance:
                    formatted_result += f"Severity {severity} finding (out of 5):\n\n"
                    formatted_result += f"{result_to_format['Explanations'][i]}\n\n"
                    formatted_result += f"Recommend replacing: {result_to_format['Phrases'][i]}\n"
                    formatted_result += f"          With this: {result_to_format['Recommendations'][i]}\n\n"
        except Exception as e:
            # include exception in returned results, with raw JSON results
            formatted_result = (f"Error occurred formatting result: {str(e)}\n\n"
                                f"Raw JSON: {json.dumps(result_to_format, indent=4)}")
            self.evaluation_engine.logger.error(formatted_result)
        return formatted_result

    def standardize_result(self, result: dict | None = None) -> list[dict]:
        """
        Reorganize the evaluation result into a list of recommendations in a standardized format.

        :param result: Evaluation result to format (or None to use the evaluation_result attribute).
        :type result: dict | None
        :return: List of recommendations, each of which is a dict with the following keys: importance (int 1-5),
            replacement_original (str), replacement_suggested (str), explanation (str).
        :rtype: list[dict]
        """

        # use evaluation_result attribute if no result is passed
        if result is not None:
            result_to_format = result
        else:
            result_to_format = self.evaluation_result

        # organize and return result
        return_list = []
        try:
            # first make sure we have a result to organize
            if result_to_format is not None and result_to_format["Number of phrases"] > 0:
                # sort the results by severity
                sorted_indices = sorted(range(len(result_to_format["Severities"])),
                                        key=lambda idx: result_to_format["Severities"][idx], reverse=True)
                # run through each result, appending to our return list
                for i in sorted_indices:
                    return_list.append({
                        'importance': result_to_format['Severities'][i],
                        'replacement_original': result_to_format['Phrases'][i],
                        'replacement_suggested': result_to_format['Recommendations'][i],
                        'explanation': result_to_format['Explanations'][i]
                    })
        except Exception as e:
            raise RuntimeError(f"Error occurred formatting result: {str(e)}\n\n"
                               f"Raw JSON: {json.dumps(result_to_format, indent=4)}") from e
        return return_list


class BiasEvaluationLens(EvaluationLens):
    """
    Lens for evaluating a survey excerpt for any of the following:

    * Stereotypical representations of gender, ethnicity, origin, religion, or other social categories.
    * Distorted or biased representations of events, topics, groups, or individuals.
    * Use of discriminatory or insensitive language towards certain groups or topics.
    * Implicit or explicit assumptions made in the text or unquestioningly adopted that could be based on prejudices.
    * Prejudiced descriptions or evaluations of abilities, characteristics, or behaviors.
    """

    def __init__(self, evaluation_engine: EvaluationEngine):
        """
        Override default constructor to provide lens-specific prompt template and followup questions.

        :param evaluation_engine: EvaluationEngine instance to use for evaluation.
        :type evaluation_engine: EvaluationEngine
        """

        # Note that this prompt was constructed from the example in this blog post:
        # https://www.linkedin.com/pulse/using-chatgpt-counter-bias-prejudice-discrimination-johannes-schunter/
        lens_system_prompt_template = """You are an AI designed to evaluate questionnaires and other survey instruments
used by researchers and M&E professionals. You are an expert in survey methodology with training equivalent to a member
of the American Association for Public Opinion Research (AAPOR) with a Ph.D. in survey methodology from University of
Michigan’s Institute for Social Research. You are also an expert in the areas of gender equality, discrimination,
anti-racism, and anti-colonialism. You consider primarily the content, context, and questions provided to you, and then
content and methods from the most widely-cited academic publications and public and nonprofit research organizations.

You always give truthful, factual answers. When asked to give your response in a specific format, you always give your
answer in the exact format requested. You never give offensive responses. If you don’t know the answer to a question,
you truthfully say you don’t know.

You will be given an excerpt from a questionnaire or survey instrument between |@| and |@| delimiters. The context and
location(s) for that excerpt are as follows:

Survey context: {survey_context}

Survey locations: {survey_locations}

Assume that this survey will be administered by a trained enumerator who asks each question and reads each prompt or
instruction as indicated in the excerpt. Your job is to review the excerpt for:

a. Stereotypical representations of gender, ethnicity, origin, religion, or other social categories.
b. Distorted or biased representations of events, topics, groups, or individuals.
c. Use of discriminatory or insensitive language towards certain groups or topics.
d. Implicit or explicit assumptions made in the text or unquestioningly adopted that could be based on prejudices.
e. Prejudiced descriptions or evaluations of abilities, characteristics, or behaviors.

Respond in JSON format with all of the following fields:

* Phrases: a list containing all problematic phrases from the excerpt that you found in your review (each phrase should
be an exact quote from the excerpt)

* Number of phrases: the exact number of phrases in Phrases

* Recommendations: a list containing suggested replacement phrases, one for each of the phrases in Phrases (in the same
order as Phrases; each replacement phrase should be an exact quote that can exactly replace the corresponding phrase in
Phrases)

* Explanations: a list containing explanations for why the phrases are problematic, one for each of the phrases in
Phrases (in the same order as Phrases)

* Severities: a list containing the severity of each identified issue, one for each of the phrases in Phrases (in the
same order as Phrases); each severity should be expressed as a number on a scale from 1 for the least severe issues
(minor phrasing issues that are very unlikely to offend respondents or substantively affect their responses) to 5 for
the most severe issues (problems that are very likely to offend respondents or substantively affect responses in a way
that introduces bias and/or variance)"""

        lens_question_template = """Excerpt: |@|{survey_excerpt}|@|"""

        lens_followups = [
            {
                'condition_func': EvaluationLens.condition_list_has_less_than_elements,
                'condition_key': 'Phrases',
                'condition_value': 1,
                'prompt_template': """Are you certain that there are no problematic phrases? If appropriate, please 
respond with a revised JSON response (including all fields). If you have no changes to propose, respond with an empty
JSON response of {{}}."""
            },
            {
                'condition_func': EvaluationLens.condition_list_has_greater_or_equal_elements,
                'condition_key': 'Phrases',
                'condition_value': 1,
                'prompt_template': """Are you certain that (1) the Phrases, Recommendations, Explanations, and
Severities lists each have exactly {Number of phrases} elements, in the same parallel order; and (2) every element of
the Severities list is a 1, 2, 3, 4, or 5, depending on the severity of the issue? If appropriate, please respond with
a revised JSON response (including all fields). If you have no changes to propose, respond with an empty JSON response
of {{}}."""
            }
        ]

        # call super constructor
        super().__init__(lens_system_prompt_template, lens_question_template, lens_followups, evaluation_engine)

    @overrides
    def evaluate(self, chat_history: list = None, survey_context: str = "", survey_locations: str = "",
                 survey_excerpt: str = "", **kwargs) -> dict:
        """
        Override default evaluate method.

        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :param survey_context: Information about the survey context.
        :type survey_context: str
        :param survey_locations: Information about the survey location(s).
        :type survey_locations: str
        :param survey_excerpt: Excerpt from the survey instrument to evaluate.
        :type survey_excerpt: str
        :param kwargs: Keyword arguments to use for formatting the task system prompt and question.
        :type kwargs: Any
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        return super().evaluate(chat_history=chat_history, survey_context=survey_context,
                                survey_locations=survey_locations,
                                survey_excerpt=EvaluationEngine.clean_whitespace(survey_excerpt), **kwargs)

    @overrides
    async def a_evaluate(self, chat_history: list = None, survey_context: str = "", survey_locations: str = "",
                         survey_excerpt: str = "", **kwargs) -> dict:
        """
        Override default a_evaluate method.

        :param chat_history: Chat history to use for the evaluation chain (or None for none).
        :type chat_history: list
        :param survey_context: Information about the survey context.
        :type survey_context: str
        :param survey_locations: Information about the survey location(s).
        :type survey_locations: str
        :param survey_excerpt: Excerpt from the survey instrument to evaluate.
        :type survey_excerpt: str
        :param kwargs: Keyword arguments to use for formatting the task system prompt and question.
        :type kwargs: Any
        :return: A dict with result ("success" or "error"), error (if result is "error"), response (a dict),
            and history (a list with the full history of the evaluation chain, each item of which is a list with two
            strings, a prompt and a response).
        :rtype: dict
        """

        return await super().a_evaluate(chat_history=chat_history, survey_context=survey_context,
                                        survey_locations=survey_locations,
                                        survey_excerpt=EvaluationEngine.clean_whitespace(survey_excerpt), **kwargs)

    def format_result(self, result: dict | None = None, minimum_importance: int = 0) -> str:
        """
        Format the evaluation result as a human-readable string.

        :param result: Evaluation result to format (or None to use the evaluation_result attribute).
        :type result: dict | None
        :param minimum_importance: Minimum importance score for filtering results (defaults to 0, which doesn't filter).
        :type minimum_importance: int
        :return: Formatted evaluation result.
        :rtype: str
        """

        # use evaluation_result attribute if no result is passed
        if result is not None:
            result_to_format = result
        else:
            result_to_format = self.evaluation_result

        # return empty string if no result is available to format
        if result_to_format is None:
            return ""

        # format and return result
        try:
            formatted_result = ""
            sorted_indices = sorted(range(len(result_to_format["Severities"])),
                                    key=lambda idx: result_to_format["Severities"][idx], reverse=True)
            for i in sorted_indices:
                severity = result_to_format['Severities'][i]
                # only report findings with severity greater than or equal to minimum_importance
                if severity >= minimum_importance:
                    formatted_result += f"Severity {severity} finding (out of 5):\n\n"
                    formatted_result += f"{result_to_format['Explanations'][i]}\n\n"
                    formatted_result += f"Recommend replacing: {result_to_format['Phrases'][i]}\n"
                    formatted_result += f"          With this: {result_to_format['Recommendations'][i]}\n\n"
        except Exception as e:
            # include exception in returned results, with raw JSON results
            formatted_result = (f"Error occurred formatting result: {str(e)}\n\n"
                                f"Raw JSON: {json.dumps(result_to_format, indent=4)}")
            self.evaluation_engine.logger.error(formatted_result)
        return formatted_result

    def standardize_result(self, result: dict | None = None) -> list[dict]:
        """
        Reorganize the evaluation result into a list of recommendations in a standardized format.

        :param result: Evaluation result to format (or None to use the evaluation_result attribute).
        :type result: dict | None
        :return: List of recommendations, each of which is a dict with the following keys: importance (int 1-5),
            replacement_original (str), replacement_suggested (str), explanation (str).
        :rtype: list[dict]
        """

        # use evaluation_result attribute if no result is passed
        if result is not None:
            result_to_format = result
        else:
            result_to_format = self.evaluation_result

        # organize and return result
        return_list = []
        try:
            # first make sure we have a result to organize
            if result_to_format is not None and result_to_format["Number of phrases"] > 0:
                # sort the results by severity
                sorted_indices = sorted(range(len(result_to_format["Severities"])),
                                        key=lambda idx: result_to_format["Severities"][idx], reverse=True)
                # run through each result, appending to our return list
                for i in sorted_indices:
                    return_list.append({
                        'importance': result_to_format['Severities'][i],
                        'replacement_original': result_to_format['Phrases'][i],
                        'replacement_suggested': result_to_format['Recommendations'][i],
                        'explanation': result_to_format['Explanations'][i]
                    })
        except Exception as e:
            raise RuntimeError(f"Error occurred formatting result: {str(e)}\n\n"
                               f"Raw JSON: {json.dumps(result_to_format, indent=4)}") from e
        return return_list



#  Copyright (c) 2023-24 Higher Bar AI, PBC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Utility functions for reading and parsing survey files."""

import os
import tempfile
import logging
import asyncio
import requests
from urllib.parse import urlparse

from surveyeval.html_tools import MarkdownifyHTMLProcessor
from langchain_core.pydantic_v1 import BaseModel, Field
from typing import List, Optional, TypedDict, Tuple
import uuid
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    ToolMessage,
)
from langchain_core.runnables import Runnable
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder

from langchain_openai import ChatOpenAI, AzureChatOpenAI
from langchain.chains import LLMChain
from langchain.callbacks import get_openai_callback
from typing import Callable
import nltk
import spacy
import csv
import re
import pytesseract
from pypdf import PdfReader
from tabula.io import read_pdf
from pdf2image import convert_from_path
from langchain.schema import Document as SchemaDocument
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.document_loaders import UnstructuredFileLoader, UnstructuredExcelLoader
import importlib.resources as pkg_resources
from openpyxl import load_workbook

# initialize global resources
parser_logger = logging.getLogger(__name__)
nltk.download('punkt')
nlp = spacy.load('en_core_web_sm')
empty_form_path = str(pkg_resources.files('surveyeval').joinpath('resources/EmptyForm.xlsx'))

# initialize global variables
langchain_splitter_chunk_size = 3000
langchain_splitter_overlap_size = 500

# initialize parsing schema
module_name_spec = ('The short name or identifier of the survey module, if any, which is the main section, category, '
                    'or group within which a questionnaire or digital form\'s questions or form fields are located. '
                    'Examples of module names: "demographics", "health", "hhmembers", "factors", etc.')
module_title_spec = ('The longer title of the survey module, if any. Examples of module titles: "Household '
                     'demographics", "People living in the household", "Factors that influence take-up", etc.')
module_intro_spec = ('The introductory text or instructions that appear at the start of the module, before the '
                     'module\'s first question begins. For example: "Now we\'re going to ask some questions about '
                     'the members of your household."')
questions_spec = ('The list of questions or form fields within the module, including the question ID, question text, '
                  'instructions, response options, and language.')
question_id_spec = ('The numeric or alphanumeric identifier or short variable name identifying the '
                    'question, usually located just before or at the beginning of the question.')
question_spec = ('The exact text of the question or form field, including any introductory text that provides '
                 'context or explanation. Often follows a unique question ID of some sort, like "2.01." or "gender:". '
                 'Should not include response options, which should be included in the "options" field, or extra '
                 'enumerator or interviewer instructions (including interview probes), which should be included in the '
                 '"instructions" field. Be careful: the same question might be asked in multiple languages, and each '
                 'translation should be included as a separate question with the proper language name in the '
                 '"language" field. Never translate between languages or otherwise alter the question text in any way.')
instructions_spec = ('Instructions or other guidance about how to ask or answer the '
                     'question, including enumerator or interviewer instructions. If the question includes '
                     'a list of specific response options, do NOT include those in the instructions. However, if '
                     'there is guidance as to how to fill out an open-ended numeric or text response, or guidance '
                     'about how to choose among the options, include that guidance here.')
options_spec = ("The list of specific response options for multiple-choice questions, "
                "including both the label and the internal value (if specified) for each option. For example, "
                "a 'Male' label might be coupled with an internal value of '1', 'M', or even 'male'. "
                "Separate response options with a space, three pipe symbols ('|||'), and another space, and, "
                "if there is an internal value, add a space, three # symbols ('###'), and the internal value "
                "at the end of the label. For example: 'Male ### 1 ||| Female ### 2' (codes included) or "
                "'Male ||| Female' (no codes); 'Yes ### yes ||| No ### no', 'Yes ### 1 ||| No ### 0', 'Yes ### "
                "y ||| No ### n', or 'YES ||| NO'. Do NOT include fill-in-the-blank content here, only "
                "multiple-choice options. If the question is open-ended, leave this field blank. ")
language_spec = 'The primary language in which the question text is written.'


class Question(BaseModel):
    """Information about a survey question or form field."""

    question_id: Optional[str] = Field(..., description=question_id_spec)
    question: Optional[str] = Field(..., description=question_spec)
    instructions: Optional[str] = Field(..., description=instructions_spec)
    options: Optional[str] = Field(..., description=options_spec)
    language: Optional[str] = Field(..., description=language_spec)


class Module(BaseModel):
    """Information about a survey module, which is the main section, category, or group within which a questionnaire
or digital form's questions or form fields are located. Examples of modules include: "Demographics", "Health",
"Household members", "Education", etc. All questions are located within modules, even if there is no name, title, or
introductory text for the module."""

    module_name: Optional[str] = Field(..., description=module_name_spec)
    module_title: Optional[str] = Field(..., description=module_title_spec)
    module_intro: Optional[str] = Field(..., description=module_intro_spec)
    questions: Optional[List[Question]] = Field(..., description=questions_spec)


class ModuleList(BaseModel):
    """List of extracted survey modules, including all extracted questions or form fields, from a questionnaire or
digital form."""

    modules: List[Module]


class Example(TypedDict):
    """
    A representation of an example consisting of text input and expected tool calls.

    For extraction, the tool calls are represented as instances of pydantic model.
    """

    input: str                          # the example text
    tool_calls: List[BaseModel]         # instances of pydantic model that should be extracted


prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            "You are an expert questionnaire extraction algorithm. "
            "Only extract relevant information from the text. "
            "If you do not know the value of an attribute asked "
            "to extract, return null for the attribute's value.",
        ),
        MessagesPlaceholder("examples"),
        ("human", "{text}"),
    ]
)


def tool_example_to_messages(example: Example) -> List[BaseMessage]:
    """
    Convert an example into a list of messages that can be fed into an LLM.

    This code is an adapter that converts our example to a list of messages
    that can be fed into a chat model.

    The list of messages per example corresponds to:

    1) HumanMessage: contains the content from which content should be extracted.
    2) AIMessage: contains the extracted information from the model
    3) ToolMessage: contains confirmation to the model that the model requested a tool correctly.

    The ToolMessage is required because some of the chat models are hyper-optimized for agents
    rather than for an extraction use case.
    """
    messages: List[BaseMessage] = [HumanMessage(content=example["input"])]
    openai_tool_calls = []
    for tool_call in example["tool_calls"]:
        openai_tool_calls.append(
            {
                "id": str(uuid.uuid4()),
                "type": "function",
                "function": {
                    # The name of the function right now corresponds
                    # to the name of the pydantic model
                    # This is implicit in the API right now,
                    # and will be improved over time.
                    "name": tool_call.__class__.__name__,
                    "arguments": tool_call.json(),
                },
            }
        )
    messages.append(
        AIMessage(content="", additional_kwargs={"tool_calls": openai_tool_calls})
    )
    tool_outputs = ["You have correctly called this tool."] * len(openai_tool_calls)
    for output, tool_call in zip(tool_outputs, openai_tool_calls):
        messages.append(ToolMessage(content=output, tool_call_id=tool_call["id"]))
    return messages


examples = [
    (
        """2. Demographics

We’ll begin with some questions so that we can get to know you and your family.

[BIRTHYR] What year were you born?

[GENDER] Which gender do you identify with?

Female

Male

Non-binary

Prefer not to answer

[ZIPCODE] What is your zip code?""",
        [
            Module(
                module_name=None,
                module_title="2. Demographics",
                module_intro="We’ll begin with some questions so that we can get to know you and your family.",
                questions=[
                    Question(
                        question_id="BIRTHYR",
                        question="What year were you born?",
                        instructions=None,
                        options=None,
                        language="English"
                    ),
                    Question(
                        question_id="GENDER",
                        question="Which gender do you identify with?",
                        instructions=None,
                        options="Female ||| Male ||| Non-binary ||| Prefer not to answer",
                        language="English"
                    ),
                    Question(
                        question_id="ZIPCODE",
                        question="What is your zip code?",
                        instructions=None,
                        options=None,
                        language="English"
                    ),
                ]
            ),
        ]
    ),
    (
        """[EDUCATION] What is the highest level of education you have completed?

o Less than high school

o High school / GED

o Some college

o 2-year college degree

o 4-year college degree

o Vocational training

o Graduate degree

o Prefer not to answer""",
        [
            Module(
                module_name=None,
                module_title=None,
                module_intro=None,
                questions=[
                    Question(
                        question_id="EDUCATION",
                        question="What is the highest level of education you have completed?",
                        instructions=None,
                        options="Less than high school ||| High school / GED ||| Some college ||| 2-year college "
                                "degree ||| 4-year college degree ||| Vocational training ||| Graduate degree ||| "
                                "Prefer not to answer",
                        language="English"
                    ),
                ]
            ),
        ]
    ),
    (
        """And how much do you disagree or agree with the following statements? For each statement, please """
        """rate how much the pair of traits applies to you, even if one trait applies more strongly than the """
        """other. I see myself as... [For each: 1=Strongly disagree, 7=Strongly agree]

[BIG5Q1] Extraverted, enthusiastic

[BIG5Q2] Critical, quarrelsome

[BIG5Q3] Dependable, self-disciplined""",
        [
            Module(
                module_name=None,
                module_title=None,
                module_intro=None,
                questions=[
                    Question(
                        question_id="BIG5Q1",
                        question="And how much do you disagree or agree with the following statement? I see myself "
                                 "as... Extraverted, enthusiastic",
                        instructions="Please rate how much the pair of traits applies to you, even if one trait "
                                     "applies more strongly than the other.",
                        options="Strongly disagree ### 1 ||| 2 ||| 3 ||| 4 ||| 5 ||| 6 ||| Strongly agree ### 7",
                        language="English"
                    ),
                    Question(
                        question_id="BIG5Q2",
                        question="And how much do you disagree or agree with the following statement? I see myself "
                                 "as... Critical, quarrelsome",
                        instructions="Please rate how much the pair of traits applies to you, even if one trait "
                                     "applies more strongly than the other.",
                        options="Strongly disagree ### 1 ||| 2 ||| 3 ||| 4 ||| 5 ||| 6 ||| Strongly agree ### 7",
                        language="English"
                    ),
                    Question(
                        question_id="BIG5Q3",
                        question="And how much do you disagree or agree with the following statement? I see myself "
                                 "as... Dependable, self-disciplined",
                        instructions="Please rate how much the pair of traits applies to you, even if one trait "
                                     "applies more strongly than the other.",
                        options="Strongly disagree ### 1 ||| 2 ||| 3 ||| 4 ||| 5 ||| 6 ||| Strongly agree ### 7",
                        language="English"
                    ),
                ]
            ),
        ]
    ),
    (
        """4. Savings Habits

Next, we will ask questions over your monthly saving habits and the potential methods that are used to save money.

1. On average, how much money do you spend monthly on essential goods below that contribute to your wellbeing """
        """(explain/ add in an example)

2. How do you typically spend your monthly income? (choose all that may apply)

a. Home and Housing

b. Retirement

c. Bills and Utility

d. Medical (Physical and Mental Treatment and Care)

e. Taxes

f. Insurance

g. Credit Card Payments (if applicable)

h. Food

i. Shopping and personal items

j. Other

k. I am not able to save money each month

l. Nothing

m. Don’t Know

3. Do you contribute the same amount or more to your savings each month?""",
        [
            Module(
                module_name=None,
                module_title="4. Savings Habits",
                module_intro="Next, we will ask questions over your monthly saving habits and the potential methods "
                             "that are used to save money.",
                questions=[
                    Question(
                        question_id="1",
                        question="On average, how much money do you spend monthly on essential goods below that "
                                 "contribute to your wellbeing (explain/ add in an example)",
                        instructions=None,
                        options=None,
                        language="English"
                    ),
                    Question(
                        question_id="2",
                        question="How do you typically spend your monthly income?",
                        instructions=None,
                        options="Home and Housing ### a ||| Retirement ### b ||| Bills and Utility ### c ||| Medical "
                                "(Physical and Mental Treatment and Care) ### d ||| Taxes ### e ||| Insurance ### f "
                                "||| Credit Card Payments (if applicable) ### g ||| Food ### h ||| Shopping and "
                                "personal items ### i ||| Other ### j ||| I am not able to save money each month ### "
                                "k ||| Nothing ### l ||| Don’t Know ### m",
                        language="English"
                    ),
                    Question(
                        question_id="3",
                        question="Do you contribute the same amount or more to your savings each month?",
                        instructions=None,
                        options=None,
                        language="English"
                    ),
                ]
            ),
        ]
    ),
    (
        """<h1>Round 1, June 2020 Eng</h1>
<table border="1" class="dataframe">
<tbody>
<tr>
<td>Module</td>
<td>Section</td>
<td>Variable</td>
<td>Question</td>
<td>Response set</td>
</tr>
<tr>
<td></td>
<td>CONS. Introduction and Consent</td>
<td></td>
<td></td>
<td></td>
</tr>
<tr>
<td>core</td>
<td>CONS</td>
<td></td>
<td>Good morning/afternoon/evening. My name is ______________________ from Innovations from Poverty Action, a """
        """Mexican research NGO. \n \n We would like to invite you to participate in a survey lasting about """
        """20 minutes about the effects of covid-19 on economic and social conditions in the Mexico City """
        """metropolitan area. If you are eligible for the survey we will compensate you [30 pesos] in """
        """airtime for completing your survey.</td>
<td></td>
</tr>
<tr>
<td></td>
<td>CONS</td>
<td>cons1</td>
<td>Can I give you more information?</td>
<td>Y/N</td>
</tr>
<tr>
<td></td>
<td>CONS</td>
<td></td>
<td>*If cons1=N\n Thank you for your response. We will end the survey now.</td>
<td>[End survey]</td>
</tr>
<tr>
<td>core</td>
<td>END</td>
<td>end4</td>
<td>What is your first name?</td>
<td></td>
</tr>
<tr>
<td>core</td>
<td>DEM</td>
<td>dem1</td>
<td>How old are you?</td>
<td>*Enter age*\n ###</td>
</tr>
<tr>
<td>core</td>
<td>CONS</td>
<td></td>
<td>*If DEM1&lt;18*\n Thank you for your response. We will end the survey now.</td>
<td>[End survey]</td>
</tr>""",
        [
            Module(
                module_name="CONS",
                module_title="Introduction and Consent",
                module_intro="Good morning/afternoon/evening. My name is ______________________ from "
                             "Innovations from Poverty Action, a "
                             "Mexican research NGO. \n \n We would like to invite you to participate in a survey "
                             "lasting about 20 minutes about the effects of covid-19 on economic and social "
                             "conditions in the Mexico City metropolitan area. If you are eligible for the survey "
                             "we will compensate you [30 pesos] in airtime for completing your survey.",
                questions=[
                    Question(
                        question_id="cons1",
                        question="Can I give you more information?",
                        instructions="*If cons1=N\n Thank you for your response. We will end the survey now. "
                                     "[End survey]",
                        options="Y ||| N",
                        language="English"
                    ),
                    Question(
                        question_id="end4",
                        question="What is your first name?",
                        instructions=None,
                        options=None,
                        language="English"
                    ),
                    Question(
                        question_id="dem1",
                        question="How old are you?",
                        instructions="*Enter age*\n ###\n"
                                     "*If DEM1&lt;18*\n Thank you for your response. We will end the survey now. "
                                     "[End survey]",
                        options=None,
                        language="English"
                    ),
                ]
            ),
        ]
    ),
    (
        """<tr>
<td>MEXICO</td>
<td>INC</td>
<td>inc11_mex</td>
<td>*If YES to INC12_mex*\n If schools and daycares remained closed and workplaces re-opened, would anyone in your """
        """household have to stay home and not return to work in order to care for children too young to """
        """stay at home without supervision?</td>
<td>*Read out, select multiple possible*\n Grandparents\n Hired babysitter\n Neighbors\n Mother who normally st\n """
        """Mother who normally works outside the home\n Father who normally works outside the home\n Older """
        """sibling\n DNK</td>
</tr>
<tr>
<td></td>
<td>NET. Social Safety Net</td>
<td></td>
<td></td>
<td></td>
</tr>
<tr>
<td>core</td>
<td>NET</td>
<td>net1</td>
<td>Do you usually receive a regular transfer from any cash transfer or other in-kind social support program?\n """
        """\n HINT: Social safety net programs include cash transfers and in-kind food transfers (food """
        """stamps and vouchers, food rations, and emergency food distribution). Example includes XXX cash """
        """transfer programme.</td>
<td>Y/N/DNK</td>
</tr>""",
        [
            Module(
                module_name=None,
                module_title=None,
                module_intro=None,
                questions=[
                    Question(
                        question_id="inc11_mex",
                        question="If schools and daycares remained closed and workplaces re-opened, would "
                                 "anyone in your household have to stay home and not return to work in order to care "
                                 "for children too young to stay at home without supervision?",
                        instructions="*If YES to INC12_mex*\n*Read out, select multiple possible*",
                        options="Grandparents ||| Hired babysitter ||| Neighbors ||| Mother who normally st ||| Mother "
                                "who normally works outside the home ||| Father who normally works outside the home "
                                "||| Older sibling ||| DNK",
                        language="English"
                    ),
                ]
            ),
            Module(
                module_name="NET",
                module_title="Social Safety Net",
                module_intro=None,
                questions=[
                    Question(
                        question_id="net1",
                        question="Do you usually receive a regular transfer from any cash transfer or other "
                                 "in-kind social support program?\n \n HINT: Social safety net programs "
                                 "include cash transfers and in-kind food transfers (food stamps and vouchers, "
                                 "food rations, and emergency food distribution). Example includes XXX cash "
                                 "transfer programme.",
                        instructions="*If cons1=N\n Thank you for your response. We will end the survey now. "
                                     "[End survey]",
                        options="Y ||| N ||| DNK",
                        language="English"
                    ),
                ]
            ),
        ]
    ),
    (
        """<tr>
<td></td>
<td>POL. POLICING</td>
<td></td>
<td></td>
<td></td>
</tr>
<tr>
<td>MEXICO</td>
<td>POL</td>
<td></td>
<td>Now I am going to ask you some questions about the main problems of insecurity in Mexico City and the """
        """performance of the city police since the coronavirus pandemic began around March 20, 2020.</td>
<td></td>
</tr>
<tr>
<td>MEXICO</td>
<td>POL</td>
<td>POL1</td>
<td>Compared to the level of insecurity that existed in your neighborhood before the pandemic began, do you """
        """consider that the level of insecurity in your neighborhood decreased, remained more or less the """
        """same, or increased?</td>
<td>Decreased\n it was more or less the same\n increased\n (777) Doesn’t answer\n (888) Doesn’t know\n (999) """
        """Doesn’t apply</td>
</tr>""",
        [
            Module(
                module_name="POL",
                module_title="POLICING",
                module_intro="Now I am going to ask you some questions about the main problems of insecurity in Mexico "
                             "City and the performance of the city police since the coronavirus pandemic began around "
                             "March 20, 2020.",
                questions=[
                    Question(
                        question_id="POL1",
                        question="Compared to the level of insecurity that existed in your neighborhood before "
                                 "the pandemic began, do you consider that the level of insecurity in your "
                                 "neighborhood decreased, remained more or less the same, or increased?",
                        instructions=None,
                        options="Decreased ||| it was more or less the same ||| increased ||| Doesn’t answer ### 777 "
                                "||| Doesn’t know ### 888 ||| Doesn’t apply ### 999",
                        language="English"
                    ),
                ]
            ),
        ]
    ),
    (
        """Module: HISTORY1 | Question: 1408 | Language: AFRIKAANS

Wanneer was die laaste keer dat jy ’n Papsmeer gehad het?

WITHIN THE LAST 3 YEARS = 1

4-5 YEARS AGO = 2

6-10 YEARS AGO = 3

MORE THAN 10 YEARS AGO = 4

DON'T KNOW/DON'T REMEMBER = 8""",
        [
            Module(
                module_name="HISTORY1",
                module_title=None,
                module_intro=None,
                questions=[
                    Question(
                        question_id="1408",
                        question="Wanneer was die laaste keer dat jy ’n Papsmeer gehad het?",
                        instructions=None,
                        options="WITHIN THE LAST 3 YEARS ### 1 ||| 4-5 YEARS AGO ### 2 ||| 6-10 YEARS AGO ### 3 ||| "
                                "MORE THAN 10 YEARS AGO ### 4 ||| DON'T KNOW/DON'T REMEMBER ### 8",
                        language="AFRIKAANS"
                    ),
                ]
            ),
        ]
    ),
    (
        """Module: FACTORS | Question: 1401a | Language: English

Many different factors can prevent women from getting medical advice or treatment for themselves. When you are """
        """sick and want to get medical advice or treatment, is the following a big problem or not a big problem: """
        """Getting permission to go to the doctor?

BIG PROBLEM = 1

NOT A BIG PROBLEM = 2""",
        [
            Module(
                module_name="FACTORS",
                module_title=None,
                module_intro=None,
                questions=[
                    Question(
                        question_id="1401a",
                        question="Many different factors can prevent women from getting medical advice or treatment "
                                 "for themselves. When you are sick and want to get medical advice or treatment, is "
                                 "the following a big problem or not a big problem: Getting permission to go to the "
                                 "doctor?",
                        instructions=None,
                        options="BIG PROBLEM ### 1 ||| NOT A BIG PROBLEM ### 2",
                        language="English"
                    ),
                ]
            ),
        ]
    ),
    (
        """Module: HISTORY3 | Question: 1454 | Language: AFRIKAANS

Hoeveel keer het dit in die afgelope 12 maande met jou gebeur?

NUMBER OF TIMES: _______""",
        [
            Module(
                module_name="HISTORY3",
                module_title=None,
                module_intro=None,
                questions=[
                    Question(
                        question_id="1454",
                        question="Hoeveel keer het dit in die afgelope 12 maande met jou gebeur?",
                        instructions="NUMBER OF TIMES:",
                        options=None,
                        language="AFRIKAANS"
                    ),
                ]
            ),
        ]
    ),
]


def set_langchain_splits(chunk_size: int, overlap_size: int):
    """
    Set the chunk size and overlap size for the (default) langchain splitter.

    :param chunk_size: Size of each chunk.
    :type chunk_size: int
    :param overlap_size: Size of the overlap.
    :type overlap_size: int
    """

    global langchain_splitter_chunk_size, langchain_splitter_overlap_size
    langchain_splitter_chunk_size = chunk_size
    langchain_splitter_overlap_size = overlap_size


def clean_whitespace(s: str) -> str:
    """
    Clean and standardize whitespace in a string.

    This includes converting tabs to spaces, trimming leading/trailing whitespace from each line,
    reducing multiple spaces to single ones, and reducing more than three consecutive linebreaks to three.

    :param s: String to clean.
    :type s: str
    :return: Cleaned string.
    :rtype: str
    """

    # convert tabs to spaces
    s = re.sub(r'\t', ' ', s)

    # trim leading and trailing whitespace from each line
    lines = s.split('\n')
    trimmed_lines = [line.strip() for line in lines]
    s = '\n'.join(trimmed_lines)

    # reduce multiple spaces to a single space
    s = re.sub(' +', ' ', s)

    # reduce more than two consecutive linebreaks to two
    s = re.sub(r'\n{3,}', '\n\n', s)

    return s


def split_langchain(content, create_doc: bool = True) -> list:
    """
    Split content into chunks using langchain.

    :param content: Content to split.
    :param create_doc: Flag to indicate whether to create a langchain Document.
    :type create_doc: bool
    :return: List of langchain documents.
    :rtype: list
    """

    # split content into chunks, with overlap to ensure that we capture entire questions (at risk of duplication)
    doc = SchemaDocument(page_content=content) if create_doc else content
    return RecursiveCharacterTextSplitter(chunk_size=langchain_splitter_chunk_size,
                                          chunk_overlap=langchain_splitter_overlap_size).split_documents([doc])


def split_nltk(content) -> list:
    """
    Split content into sentences using NLTK.

    :param content: Content to split.
    :return: List of sentences.
    :rtype: list
    """

    return nltk.sent_tokenize(content)


def split_spacy(content) -> list:
    """
    Split content into sentences using spaCy.

    :param content: Content to split.
    :return: List of sentences.
    :rtype: list
    """

    doc = nlp(content)
    return [sent.text for sent in doc.sents]


def split(content, splitter: Callable = split_langchain, create_doc: bool = True) -> list:
    """
    Split content using a specified splitting function.

    :param content: Content to split.
    :param splitter: Function to use for splitting. Defaults to split_langchain.
    :type splitter: Callable
    :param create_doc: Flag to indicate whether to create a langchain Document for splitting.
    :type create_doc: bool
    :return: List of split content.
    :rtype: list
    """

    split_docs = splitter(content, create_doc)
    return [doc.page_content if not isinstance(doc, str) else doc for doc in split_docs]


def read_docx(file_path: str, splitter: Callable = split_langchain) -> list:
    """
    Parse a DOCX file into a list of content chunks using a specified splitting function.

    :param file_path: Path to the DOCX file.
    :type file_path: str
    :param splitter: Function to use for splitting the content. Defaults to split_langchain.
    :type splitter: Callable
    :return: List of split content chunks.
    :rtype: list
    """

    # use langchain/unstructured to parse the DOCX file
    loader = UnstructuredFileLoader(file_path, mode="elements")
    data = loader.load()

    # concatenate page contents, using double-linebreaks as page/element separators
    content = '\n\n'.join([page.page_content for page in data])
    content = clean_whitespace(content)

    return split(content, splitter)


def parse_csv(input_csv_file: str, splitter: Callable = split_langchain) -> dict | list:
    """
    Parse a CSV file into a dictionary with questionnaire data.

    The function handles REDCap data dictionaries explicitly, then falls back to generic handling of other CSV files.

    :param input_csv_file: Path to the CSV file.
    :type input_csv_file: str
    :param splitter: Function to use for splitting in case not a REDCap data dictionary. Defaults to split_langchain.
    :type splitter: Callable
    :return: Dictionary with form data (if REDCap), otherwise list with split content.
    :rtype: dict | list
    """

    with open(input_csv_file, newline='', encoding='utf-8') as csvfile:
        reader = csv.DictReader(csvfile)

        # assume REDCap data dictionary format if "Field Type" column is present
        if 'Field Type' in reader.fieldnames:
            # process REDCap data dictionary
            all_questions = {}
            for row in reader:
                assert isinstance(row, dict)  # (type hint for IDE)
                field_type = row['Field Type']
                # only read supported field types
                if field_type in ['descriptive', 'text', 'number', 'radio', 'checkboxes', 'dropdown', 'slider',
                                  'yesno', 'truefalse']:
                    question_id = row['Variable / Field Name']
                    options = []
                    if field_type in ['radio', 'checkboxes', 'dropdown']:
                        # parse choices from "value, label | value, label | ..." pairs
                        choice_strs = row.get('Choices, Calculations, OR Slider Labels', '').split('|')
                        for choice_str in choice_strs:
                            value, label = choice_str.split(',', 1)
                            value = value.strip()
                            label = label.strip()
                            options.append({'label': label, 'value': value})
                    elif field_type == 'slider':
                        # parse choices from "label | label | ..." format
                        slider_strs = row.get('Choices, Calculations, OR Slider Labels', '').split('|')
                        for slider_str in slider_strs:
                            options.append({'label': slider_str, 'value': slider_str})
                    elif field_type == 'yesno':
                        options.append({'label': "yes", 'value': "1"})
                        options.append({'label': "no", 'value': "0"})
                    elif field_type == 'truefalse':
                        options.append({'label': "true", 'value': "1"})
                        options.append({'label': "false", 'value': "0"})

                    all_questions[question_id] = [
                        {
                            'question': row['Field Label'],
                            'instructions': row.get('Field Note', ''),
                            'options': options,
                            'language': "Unknown"   # (language not specified in REDCap data dictionary)
                        }
                    ]

                # return all questions in a single module
                form_data = {
                    "REDCap_module": {
                        "module_name": "REDCap_module",
                        "module_title": "",
                        "module_intro": "",
                        "questions": all_questions
                    }
                }
                return form_data
        else:
            # fall back to generic CSV processing
            content_list = ['\t'.join(row) for row in reader]
            content_str = '\n'.join(content_list)
            content_str = clean_whitespace(content_str)
            return split(content_str, splitter)


def convert_tables(md_text: str) -> str:
    """
    Convert Markdown tables to plain text format.

    :param md_text: Markdown text containing tables.
    :type md_text: str
    :return: Plain text representation of tables.
    :rtype: str
    """

    table_blocks = re.findall(r'\|[\s\S]+?\|', md_text)
    plain_text_tables = ""
    for table_block in table_blocks:
        rows = table_block.strip().split('\n')
        plain_text_tables += '\n'.join(
            ['\t'.join(cell.strip() for cell in row.split('|')[1:-1]) for row in rows]) + '\n\n'

    return plain_text_tables


def parse_xlsx(file_path: str, splitter: Callable = split_langchain) -> list[str] | dict:
    """
    Parse an XLSX file into a dictionary with questionnaire data.

    This function starts by assuming that the XLSX file is in XLSForm format, then falls back to generic parsing
    in the case where XLSForm parsing fails.

    :param file_path: Path to the XLSX file.
    :type file_path: str
    :param splitter: Function used to split content. Defaults to split_langchain.
    :type splitter: Callable
    :return: A dict with structured content if it was an XLSForm, otherwise a list of split content to process.
    :rtype: list[str] | dict
    """

    # first load the workbook
    wb = load_workbook(file_path)

    # if there are survey, choices, and settings worksheets, assume it's an XLSForm
    if 'survey' in wb.sheetnames and 'choices' in wb.sheetnames and 'settings' in wb.sheetnames:
        # process the XLSForm
        survey_ws = wb['survey']
        survey_columns = _get_columns_from_headers(survey_ws)
        survey_data = [row for row in survey_ws.values][1:]
        choices_ws = wb['choices']
        choices_columns = _get_columns_from_headers(choices_ws)
        choices_data = [row for row in choices_ws.values][1:]
        settings_ws = wb['settings']
        settings_columns = _get_columns_from_headers(settings_ws)

        # hack for naming flexibility: if there's a "name" column but no "value" column, rename "name" to "value"
        if 'name' in choices_columns and 'value' not in choices_columns:
            choices_columns['value'] = choices_columns['name']
            del choices_columns['name']

        # read the default_language from row 2 of the settings worksheet
        default_language = settings_ws.cell(row=2, column=settings_columns['default_language']+1).value

        # zip through the survey sheet to create a list of other languages in the form
        other_languages = []
        for column in survey_columns:
            match = re.match(r"^label:(.*)", column)
            if match:
                language = match.group(1).strip()
                # only consider adding languages we haven't already added (no dups wanted)
                if language not in other_languages:
                    # only add languages that also have translations in the choices sheet
                    if f"label:{language}" in choices_columns:
                        other_languages.append(language)

        # next, zip through the survey sheet to create a list of survey items
        output_data = {}
        group_stack = []
        groupless_count = 0
        last_group = ""
        for row in survey_data:
            if row[survey_columns['type']] and row[survey_columns['name']]:
                type_ = str(row[survey_columns['type']]).strip() if row[survey_columns['type']] is not None else ""
                name = str(row[survey_columns['name']]).strip() if row[survey_columns['name']] is not None else ""
                label = str(row[survey_columns['label']]).strip() if row[survey_columns['label']] is not None else ""
                hint = str(row[survey_columns['hint']]).strip() if row[survey_columns['hint']] is not None else ""
                
                if type_ == "begin group" or type_ == "begin repeat":
                    # new group, add to survey_data
                    output_data[name] = {
                        "module_name": name,
                        "module_title": label,
                        "module_intro": "",
                        "questions": {}
                    }
                    # and remember that we're in this group
                    group_stack.append(name)
                elif type_ == "end group" or type_ == "end repeat":
                    # pop the group stack to exit current group
                    group_stack.pop()
                elif label and type_ not in ["calculate", "calculate_here", "start", "end", "deviceid",
                                             "simserial", "phonenumber", "subscriberid", "caseid", "audio audit",
                                             "text audit", "speed violations count", "speed violations list",
                                             "speed violations audit"]:
                    # if question has a label isn't an invisible type, figure out which module it belongs to
                    if group_stack:
                        # group is most recent item added to the stack
                        module = group_stack[-1]
                    elif last_group and last_group.startswith("nomodule_"):
                        # we have a groupless module open and going, so continue with that
                        module = last_group
                    else:
                        # start a new groupless module
                        groupless_count += 1
                        module = f"nomodule_{groupless_count}"
                        # add new module to survey_data
                        output_data[module] = {
                            "module_name": module,
                            "module_title": "",
                            "module_intro": "",
                            "questions": {}
                        }
                    # remember this as the last group we were in
                    last_group = module

                    # assemble dictionary of questions
                    questions = {name: []}

                    # add item for base language first
                    item = {
                        "question": label,
                        "language": default_language,
                        "options": [],
                        "instructions": hint
                    }
                    # if the type is "select_one" or "select_multiple", add the choices
                    match = re.match(r"^(select_one|select_multiple) (.+)$", type_)
                    if match:
                        list_name = match.group(2)
                        # loop through all rows on the choices sheet with a matching list_name
                        for choice in choices_data:
                            if choice[choices_columns['list_name']] == list_name:
                                # add options one by one
                                item["options"].append({
                                    "label": str(choice[choices_columns['label']]).strip(),
                                    "value": str(choice[choices_columns['value']]).strip()
                                })
                    questions[name].append(item)

                    # next add translations (if any)
                    if other_languages:
                        for language in other_languages:
                            translated_label = str(row[survey_columns[f"label:{language}"]]).strip() \
                                if (f"label:{language}" in survey_columns
                                    and row[survey_columns[f"label:{language}"]] is not None) else ""
                            # only add translations with labels
                            if translated_label:
                                item = {
                                    "question": translated_label,
                                    "language": language,
                                    "options": [],
                                    "instructions": str(row[survey_columns[f"hint:{language}"]]).strip()
                                    if f"hint:{language}" in survey_columns
                                       and row[survey_columns[f"hint:{language}"]] is not None else ""
                                }
                                # if the type is "select_one" or "select_multiple", add the choices
                                match = re.match(r"^(select_one|select_multiple) (.+)$", type_)
                                if match:
                                    list_name = match.group(2)
                                    # loop through all rows on the choices sheet with a matching list_name
                                    for choice in choices_data:
                                        if choice[choices_columns['list_name']] == list_name:
                                            # add options one by one
                                            choice_label = str(choice[choices_columns[f"label:{language}"]]).strip() \
                                                if (f"label:{language}" in choices_columns and
                                                    choice[choices_columns[f"label:{language}"]] is not None) else ""
                                            if not choice_label:
                                                # use default language label if translated label missing
                                                choice_label = str(choice[choices_columns['label']]).strip()
                                            item["options"].append({
                                                "label": choice_label,
                                                "value": str(choice[choices_columns['value']]).strip()
                                                if f"hint:{language}" in choices_columns
                                                   and choice[choices_columns['value']] is not None else ""
                                            })
                                questions[name].append(item)

                    # update our output data with the new questions (translations for the current field)
                    output_data[module]["questions"].update(questions)

        # return structured data
        return clean_data(output_data)

    # otherwise, fallback to unstructured processing
    loader = UnstructuredExcelLoader(file_path, mode="elements")
    data = loader.load()
    content_list = []
    for page in data:
        if 'text_as_html' in page.metadata:
            html_content = page.metadata['text_as_html']
        else:
            html_content = page.page_content
        if 'page_name' in page.metadata:
            page_name = page.metadata['page_name']
            content_list.append('<h1>' + page_name + '</h1>\n' + html_content)

    # split and return the processed content
    split_content = []
    for content in content_list:
        split_content.extend(split(clean_whitespace(content), splitter))
    return split_content


def read_local_html(path: str, splitter: Callable = split_langchain) -> list:
    """
    Read and process local HTML file into a structured format.

    This function reads an HTML file from a local path, converts it to markdown,
    and then splits it into a structured format using a specified splitter function.

    :param path: Path to the local HTML file.
    :type path: str
    :param splitter: Function to split the processed markdown into a structured format.
    :type splitter: Callable
    :return: List of processed and split content.
    """

    with open(path, 'r', encoding='utf-8') as file:
        doc = SchemaDocument(page_content=file.read())
        md = MarkdownifyHTMLProcessor().process(doc)
        return splitter(md, False)


def read_html(url: str, splitter: Callable = split_langchain) -> list:
    """
    Read and process HTML content from a URL into a structured format.

    This function fetches HTML content from a given URL, converts it to markdown,
    and then splits it into a structured format using a specified splitter function.

    :param url: URL of the HTML page to be read.
    :type url: str
    :param splitter: Function to split the processed markdown into a structured format.
    :type splitter: Callable
    :return: List of processed and split content.
    :rtype: list
    """

    response = requests.get(url)
    doc = SchemaDocument(page_content=response.text)
    md = MarkdownifyHTMLProcessor().process(doc)
    return splitter(md, False)


def read_pdf_pypdf(path: str, splitter: Callable = split_langchain, split_content: bool = True) -> list:
    """
    Read and process PDF content into a structured format using PyPDF.

    This function reads a PDF file, extracts text from each page, and optionally
    splits it into a structured format using a specified splitter function.

    :param path: Path to the PDF file.
    :type path: str
    :param splitter: Function to split the extracted text into a structured format.
    :type splitter: Callable
    :param split_content: Boolean flag to determine whether to split the content or not.
    :type split_content: bool
    :return: List of processed (and optionally split) content.
    :rtype: list
    """

    content = []
    reader = PdfReader(path)
    for idx, page in enumerate(reader.pages):
        text = page.extract_text()
        content.append(text)
    return splitter('\n\n'.join(content)) if split_content else content


def extract_text_from_pdf_tabula(file_path: str, splitter: Callable = split_langchain,
                                 split_content: bool = True) -> list:
    """
    Extract and process text from a PDF file using Tabula for table extraction.

    This function reads a PDF file and extracts text from tables on each page using Tabula. It then cleans the
    extracted tables by dropping empty columns, filling NaN values, and combining them into a text format.
    Optionally, the extracted text can be split into a structured format using a specified splitter function.

    :param file_path: Path to the PDF file.
    :type file_path: str
    :param splitter: Function to split the extracted text into a structured format.
    :type splitter: Callable
    :param split_content: Boolean flag to determine whether to split the content or not.
    :type split_content: bool
    :return: List of processed (and optionally split) content from the PDF tables.
    :rtype: list
    """

    with open(file_path, 'rb') as file:
        reader = PdfReader(file)
        num_pages = len(reader.pages)

    content = []
    for page in range(1, num_pages + 1):
        # read PDF tables with Tabula
        df = read_pdf(file_path, pages=page, multiple_tables=True)

        # clean and process tables
        cleaned_tables = clean_tables(df)

        # convert cleaned tables to text
        text = convert_tables_to_text(cleaned_tables)
        content.append(text)

    return splitter('\n\n'.join(content)) if split_content else content


def clean_tables(dataframes: list) -> list:
    """
    Clean and process extracted tables from a PDF.

    :param dataframes: List of DataFrame objects representing extracted tables.
    :type dataframes: list
    :return: List of cleaned and processed DataFrame objects.
    :rtype: list
    """

    cleaned_tables = []
    for table in dataframes:
        # drop columns with all NaN values and replace NaN with empty string
        table.dropna(axis=1, how='all', inplace=True)
        table.fillna('', inplace=True)
        # append non-empty tables to the list
        if not table.empty:
            cleaned_tables.append(table)
    return cleaned_tables


def convert_tables_to_text(tables: list) -> str:
    """
    Convert cleaned tables to a text format.

    :param tables: List of cleaned DataFrame objects representing tables.
    :type tables: list
    :return: String representation of the tables.
    :rtype: str
    """

    text = ''
    for table in tables:
        # convert DataFrame to string and replace multiple spaces
        table_string = table.to_string(index=False, header=True)
        table_string = re.sub(' +', ' ', table_string)
        text += table_string + '\n\n\n'
    return text


def read_ocr(path: str, splitter: Callable = split_langchain) -> list:
    """
    Read and process text from an image-based PDF file using OCR.

    This function converts each page of a PDF to an image and then uses OCR to extract text.
    The extracted text is then split into a structured format using a specified splitter function.

    :param path: Path to the PDF file.
    :type path: str
    :param splitter: Function to split the extracted text into a structured format.
    :type splitter: Callable
    :return: List of processed and split content.
    :rtype: list
    """

    pages = convert_from_path(path)
    content = ''
    for page in pages:
        text = pytesseract.image_to_string(page)
        content += text
    return splitter(content)


def read_pdf_combined(path: str, splitter: Callable = split_langchain, min_length: int = 600) -> list:
    """
    Read and process text from a PDF file combining PyPDF and Tabula extraction methods.

    This function extracts text from a PDF using both PyPDF and Tabula methods. The combined
    text is then checked for a minimum length. If the length requirement is met, the text is
    split into a structured format using a specified splitter function. If not, it falls back
    to OCR.

    :param path: Path to the PDF file.
    :type path: str
    :param splitter: Function to split the extracted text into a structured format.
    :type splitter: Callable
    :param min_length: Minimum expected length for the combined text.
    :type min_length: int
    :return: List of processed and split content, or content from OCR if length requirement is not met.
    :rtype: list
    """

    text_pypdf_per_page = read_pdf_pypdf(path, split_langchain, False)
    text_tabula_per_page = extract_text_from_pdf_tabula(path, split_langchain, False)

    combined_text = combine_texts(text_pypdf_per_page, text_tabula_per_page)

    # clean whitespace and check for minimum length
    combined_text = clean_whitespace(combined_text)
    if len(combined_text.strip()) >= min_length:
        return splitter(combined_text)

    return read_ocr(path, splitter)


def combine_texts(texts1: list, texts2: list) -> str:
    """
    Combine texts from two lists, appending text from the second list to the first.

    :param texts1: List of strings from the first source.
    :type texts1: list
    :param texts2: List of strings from the second source.
    :type texts2: list
    :return: Combined text as a single string.
    :rtype: str
    """

    combined_text = ""
    for text1, text2 in zip(texts1, texts2):
        combined_text += text1 + "\n" + text2 + "\n"
    return combined_text


def generate_extractor_chain(model_input: str, api_base: str, openai_api_key: str, open_api_version: str,
                             provider: str = "azure") -> Runnable:
    """
    Generate an extractor chain based on the specified language model and API settings.

    :param model_input: Name of the language model to use.
    :type model_input: str
    :param api_base: Base URL for the API.
    :type api_base: str
    :param openai_api_key: API key for accessing OpenAI services.
    :type openai_api_key: str
    :param open_api_version: Version of the OpenAI API to use.
    :type open_api_version: str
    :param provider: Provider for the LLM service ("openai" for direct OpenAI, "azure" for Azure). Default is "azure".
    :type provider: str
    :return: An extraction chain configured with the specified parameters.
    :rtype: Runnable
    """

    # initialize LLM
    if provider == "azure":
        llm = AzureChatOpenAI(
            temperature=0,
            verbose=True,
            model_name=model_input,
            azure_endpoint=api_base,
            deployment_name=model_input,
            openai_api_version=open_api_version,
            openai_api_key=openai_api_key,
            openai_api_type="azure",
        )
    elif provider == "openai":
        llm = ChatOpenAI(
            temperature=0,
            verbose=True,
            model_name=model_input,
            openai_api_key=openai_api_key
        )
    else:
        raise ValueError("Unsupported provider specified. Choose 'openai' or 'azure'.")

    runnable = prompt | llm.with_structured_output(
        schema=ModuleList,
        method="function_calling",
        include_raw=False,
    )

    return runnable


def uri_validator(url: str) -> bool:
    """
    Validate if the given string is a valid URI.

    :param url: The string to validate as URI.
    :type url: str
    :return: True if the string is a valid URI, False otherwise.
    :rtype: bool
    """

    result = urlparse(url)
    return all([result.scheme, result.netloc])


def get_data_from_url(url: str, splitter: Callable = split_langchain) -> list[str] | dict:
    """
    Fetch and process data from a given URL based on file extension.

    :param url: URL of the file to process.
    :type url: str
    :param splitter: Function or method used for processing the data.
    :type splitter: Callable
    :return: A list of strings for raw text content, or a dictionary for structured content (needs no extra parsing).
    :rtype: list[str] | dict | None
    """

    valid_url = uri_validator(url)
    response = []
    if valid_url:
        response = requests.get(url.strip())
        if response.status_code != 200:
            parser_logger.log(logging.ERROR, f'Error: {response.status_code}')
            return []

    extension = url.split('.')[-1].strip().lower()
    process_functions = {
        'pdf': read_pdf_combined,
        'docx': read_docx,
        'csv': parse_csv,
        'xlsx': parse_xlsx,
        'html': read_local_html,
        'txt': read_docx
    }

    if extension in process_functions:
        # process based on the extension
        process_function = process_functions.get(extension, read_html)
        if valid_url:
            # if the URL is valid, save the file locally
            temp_file_handle, temp_file_path = tempfile.mkstemp(suffix='.' + extension)
            with os.fdopen(temp_file_handle, 'wb') as temp_file:
                temp_file.write(response.content)
            file_to_process = temp_file_path
        else:
            file_to_process = url
        retval = process_function(file_to_process, splitter)
    else:
        # for other extensions, assume it's a URL to an HTML page
        retval = read_html(url, splitter)

    # if the result is a list of documents, extract the page content for a list of strings
    if isinstance(retval, list) and retval:
        retval = [doc.page_content if not isinstance(doc, str) else doc for doc in retval]
    return retval


def total_string_length(d: dict) -> int:
    """
    Calculate the total string length of all values in a dictionary.

    :param d: The dictionary whose values' string lengths are to be summed.
    :type d: dict
    :return: The total string length of all values.
    :rtype: int
    """

    # sum length of all strings, including those in nested lists of dicts
    total_length = 0
    for value in d.values():
        if isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    total_length += total_string_length(item)
        else:
            total_length += len(str(value))

    return total_length


def clean_data(data: dict) -> dict:
    """
    Clean the provided data by removing empty modules and questions — and by removing duplicate questions,
    keeping the ones with more content.

    :param data: The data to be cleaned.
    :type data: dict
    :return: The cleaned data.
    :rtype: dict
    """

    cleaned_data = {}
    # run through all modules
    for key, dt in data.items():
        # assemble module questions, dropping empty questions
        questions = {question: question_data for question, question_data in dt['questions'].items()
                     if question and question_data}

        # skip empty modules
        if not questions:
            parser_logger.log(logging.INFO, f"Skipping empty module: {key}")
        else:
            parser_logger.log(logging.INFO, f"* Module: {key}")
            # for each question, look for questions with the same language
            for question, question_data in questions.items():
                if question_data:
                    to_remove = set()
                    for i in range(len(question_data)):
                        for j in range(i + 1, len(question_data)):
                            if question_data[i]['language'] == question_data[j]['language']:
                                # compare total string lengths and mark the shorter one for removal
                                length_i = total_string_length(question_data[i])
                                length_j = total_string_length(question_data[j])
                                if length_i > length_j:
                                    to_remove.add(j)
                                    parser_logger.log(logging.INFO, f"  * Dropping duplicate question with "
                                                                    f"shorter content: "
                                                                    f"{question} - {question_data[j]['language']} - "
                                                                    f"{question_data[j]['question']} "
                                                                    f"({length_j} <= {length_i})")
                                else:
                                    to_remove.add(i)
                                    parser_logger.log(logging.INFO, f"  * Dropping duplicate question with "
                                                                    f"shorter content: "
                                                                    f"{question} - {question_data[i]['language']} - "
                                                                    f"{question_data[i]['question']} "
                                                                    f"({length_i} <= {length_j})")

                    # remove duplicates after identifying them
                    for index in sorted(to_remove, reverse=True):
                        question_data.pop(index)

            # add module to cleaned data
            cleaned_data[key] = {
                "module_name": dt['module_name'],
                "module_title": dt['module_title'],
                "module_intro": dt['module_intro'],
                "questions": questions
            }

    return cleaned_data


async def _invoke_task(func: Callable, param_dict: dict) -> dict:
    """
    Asynchronously invoke a function with parameters, capturing and returning any exceptions that occur.

    :param func: Task invocation function.
    :type func: Callable
    :param param_dict: Parameter dict to pass to the function.
    :type param_dict: dict
    :return: A dict with result ("success" or "error"), error (if result is "error"), and response (a dict).
    :rtype: dict
    """

    try:
        return {"result": "success", "error": None, "response": await func(param_dict)}
    except Exception as e:
        parser_logger.error(f"Task triggered error: {str(e)}")
        return {"result": "error", "error": f"Task triggered error: {str(e)}", "response": None}


async def extract_data(chain: Runnable, url: str, replacement_examples: List[Tuple[str, List[Question]]] = None,
                       additional_examples: List[Tuple[str, List[Question]]] = None) -> dict:
    """
    Asynchronously process the content from a given URL and parse it into structured data.

    :param chain: The AI prediction chain to be used.
    :type chain: Runnable
    :param url: URL of the document to process.
    :type url: str
    :param replacement_examples: Replacement examples to use (if any).
    :type replacement_examples: List[Tuple[str, List[Question]]]
    :param additional_examples: Additional examples to use (if any).
    :type additional_examples: List[Tuple[str, List[Question]]]
    :return: A dict with modules (a dict with questions organized by module) and errors (a list of errors, if any).
    :rtype: dict
    """

    # construct example messages, including default set, plus any replacements or additional examples
    example_messages = []
    if replacement_examples is not None:
        for text, tool_call in replacement_examples:
            example_messages.extend(
                tool_example_to_messages({"input": text, "tool_calls": tool_call})
            )
    else:
        for text, tool_call in examples:
            example_messages.extend(
                tool_example_to_messages({"input": text, "tool_calls": tool_call})
            )
    if additional_examples is not None:
        for text, tool_call in additional_examples:
            example_messages.extend(
                tool_example_to_messages({"input": text, "tool_calls": tool_call})
            )

    # get data from the URL
    read_data = get_data_from_url(url)

    # if a dict is returned, that means the data was read as structured and we can return it straight away
    if isinstance(read_data, dict):
        return read_data

    # if it's an empty list, just return an empty dict
    if not read_data:
        return {}

    # otherwise, we have to process a content list
    # process list asynchronously, and track our LLM usage with an OpenAI callback
    with get_openai_callback() as cb:
        # create a list of tasks, then execute them asynchronously
        tasks = [_invoke_task(chain.ainvoke,
                              {"text": page, "examples": example_messages}) for page in read_data]
        results = await asyncio.gather(*tasks)

        # report LLM usage
        parser_logger.log(logging.INFO, f"Tokens consumed:: {cb.total_tokens}")
        parser_logger.log(logging.INFO, f"  Prompt tokens: {cb.prompt_tokens}")
        parser_logger.log(logging.INFO, f"  Completion tokens: {cb.completion_tokens}")
        parser_logger.log(logging.INFO, f"Successful Requests: {cb.successful_requests}")
        parser_logger.log(logging.INFO, f"Cost: ${cb.total_cost}")

    # organize questions by module and question ID
    errors_output = []
    modules_output = {}
    question_module = {}
    unknown_module_count = 0
    unknown_id_count = 0
    for task_result in results:
        # if the task failed, add the error to the errors list and continue
        if task_result['result'] == 'error':
            errors_output.append(task_result['error'])
            continue

        for module in task_result['response'].modules:
            # construct module name from whatever details we have, defaulting to auto-naming as necessary
            if module.module_name == module.module_title:
                module_key = module.module_name
            elif module.module_name and module.module_title:
                module_key = f"{module.module_name} - {module.module_title}"
            elif module.module_name:
                module_key = module.module_name
            elif module.module_title:
                module_key = module.module_title
            else:
                unknown_module_count += 1
                module_key = f"MODULE_{unknown_module_count}"

            # run through all questions in module
            for question in module.questions:
                # process question, if any
                if question.question and question.question.strip():
                    # get question ID, if available
                    question_id = question.question_id.strip() if question.question_id else ''
                    if not question_id:
                        # if no question ID is provided, generate a unique ID
                        unknown_id_count += 1
                        question_id = f"question_{unknown_id_count}"

                    if question_id in question_module:
                        # if we've seen the question ID before, add this version to the same module as before
                        question_list = modules_output[question_module[question_id]].setdefault('questions', {})
                    else:
                        # otherwise, add it to the current module, adding the module to the output as needed
                        if module_key not in modules_output:
                            # add module to output, ignoring module title if same as the module name
                            modules_output[module_key] = {
                                "module_name": module.module_name if module.module_name else "",
                                "module_title": module.module_title if module.module_title
                                                                       and module.module_title != module.module_name
                                                                    else "",
                                "module_intro": module.module_intro if module.module_intro else "",
                                "questions": {}
                            }
                        question_list = modules_output[module_key].setdefault('questions', {})
                        question_module[question_id] = module_key

                    # parse and organize options, if any
                    options = []
                    if question.options:
                        option_strs = [option.strip() for option in question.options.split('|||')]
                        for option_str in option_strs:
                            if '###' in option_str:
                                option_parts = option_str.split('###')
                                if len(option_parts) == 2:
                                    option_parts = [part.strip() for part in option_parts]
                                    options.append({'label': option_parts[0], 'value': option_parts[1]})
                                else:
                                    options.append({'label': option_str, 'value': option_str})
                            else:
                                options.append({'label': option_str, 'value': option_str})

                    # add question, grouped by module and question ID
                    if question_id not in question_list:
                        question_list[question_id] = []
                    question_list[question_id].append({
                        'question': question.question,
                        'language': question.language if question.language else '',
                        'options': options,
                        'instructions': question.instructions if question.instructions else '',
                    })

    # return cleaned-up version of the data, along with any errors
    return {"modules": clean_data(modules_output), "errors": errors_output}


async def extract_data_from_directory(path_to_ingest: str, chain: LLMChain) -> list:
    """
    Extract structured data from all files in a specified directory.

    :param path_to_ingest: Path to the directory containing files to process.
    :type path_to_ingest: str
    :param chain: The AI prediction chain to be used.
    :type chain: LLMChain
    :return: A list of data extraction output from all processed files.
    :rtype: list
    """

    data_list = []
    for root, dirs, files in os.walk(path_to_ingest):
        for file in files:
            if file.endswith((".DS_Store", ".db")):
                # skip system files
                continue

            file_path = os.path.join(root, file)
            data_list.append(await extract_data(chain, file_path))

    return data_list


def _add_header_to_first_empty_cell(worksheet, header):
    for cell in worksheet[1]:
        if cell.value is None or not cell.value:
            cell.value = header
            break
    else:  # no break, meaning no empty cell was found
        max_column = worksheet.max_column
        worksheet.cell(row=1, column=max_column + 1, value=header)


def _get_columns_from_headers(worksheet, index_boost: int = 0) -> dict:
    return {cell.value: i + index_boost for i, cell in enumerate(worksheet[1])
            if cell.value is not None and cell.value.strip() != ''}


def output_parsed_data_to_xlsform(data: dict, form_id: str, form_title: str, output_file: str):
    """
    Output parsed data to an XLSForm file.

    :param data: Parsed data to output.
    :type data: dict
    :param form_id: Form ID to set in the XLSForm file.
    :type form_id: str
    :param form_title: Form title to set in the XLSForm file.
    :type form_title: str
    :param output_file: Path to the output XLSForm file.
    :type output_file: str
    """

    # load the workbook
    wb = load_workbook(empty_form_path)
    survey_ws = wb['survey']
    survey_columns = _get_columns_from_headers(survey_ws, 1)
    choices_ws = wb['choices']
    choices_columns = _get_columns_from_headers(choices_ws, 1)
    settings_ws = wb['settings']
    settings_columns = _get_columns_from_headers(settings_ws, 1)

    # set the 'form_title' and 'form_id' values in row 2
    settings_ws.cell(row=2, column=settings_columns['form_id'], value=form_id)
    settings_ws.cell(row=2, column=settings_columns['form_title'], value=form_title)

    # initialize our row counters to start adding at the second row
    survey_row_counter = 2
    choices_row_counter = 2

    # iterate over each module in the data
    primary_language = ""
    for module in data.values():
        # create a safe version of the module name by replacing anything not a digit or a letter with an _
        safe_module_name = re.sub(r'\W+', '_', module['module_name']).lower()

        # open a group for the module
        survey_ws.cell(row=survey_row_counter, column=survey_columns['type'], value='begin group')
        survey_ws.cell(row=survey_row_counter, column=survey_columns['name'], value=safe_module_name)
        if module['module_title']:
            survey_ws.cell(row=survey_row_counter, column=survey_columns['label'], value=module['module_title'])
        survey_row_counter += 1

        # if there is a 'module_intro' value in the module, add it as a note field
        if module['module_intro']:
            survey_ws.cell(row=survey_row_counter, column=survey_columns['type'], value='note')
            survey_ws.cell(row=survey_row_counter, column=survey_columns['label'], value=module['module_intro'])
            survey_row_counter += 1

        # iterate over each question in the module
        for question_id, question_data in module['questions'].items():
            # create a safe version of the question ID by replacing anything not a digit or a letter with an _
            safe_question_id = re.sub(r'\W+', '_', question_id).lower()
            # if safe_question_id doesn't begin with a letter, put a 'q' on the front
            if not safe_question_id[0].isalpha():
                safe_question_id = 'q' + safe_question_id

            # iterate over each translation for the question
            has_choices = False
            for translation in question_data:
                # if we don't have our primary language yet, use the first language we find
                if not primary_language:
                    primary_language = translation['language']
                    settings_ws.cell(row=2, column=settings_columns['default_language'], value=primary_language)

                # set language suffix
                if translation['language'] and translation['language'] != primary_language:
                    language_suffix = f":{translation['language']}"
                    label_column = 'label' + language_suffix
                    hint_column = 'hint' + language_suffix
                    # also add translation columns as necessary
                    if label_column not in survey_columns:
                        _add_header_to_first_empty_cell(survey_ws, label_column)
                        survey_columns = _get_columns_from_headers(survey_ws, 1)
                    if translation['options'] and label_column not in choices_columns:
                        _add_header_to_first_empty_cell(choices_ws, label_column)
                        choices_columns = _get_columns_from_headers(choices_ws, 1)
                    if hint_column not in survey_columns:
                        _add_header_to_first_empty_cell(survey_ws, hint_column)
                        survey_columns = _get_columns_from_headers(survey_ws, 1)
                else:
                    label_column = 'label'
                    hint_column = 'hint'
                    # if there are options, we'll need to add them to the choices sheet (all at the end)
                    #   (we let the primary language govern whether there are choices)
                    if translation['options']:
                        has_choices = True

                # add the question to the survey sheet
                survey_ws.cell(row=survey_row_counter, column=survey_columns['type'],
                               value='text' if not has_choices else 'select_one ' + safe_question_id)
                survey_ws.cell(row=survey_row_counter, column=survey_columns['name'], value=safe_question_id)
                survey_ws.cell(row=survey_row_counter, column=survey_columns[label_column],
                               value=translation['question'])
                survey_ws.cell(row=survey_row_counter, column=survey_columns[hint_column],
                               value=translation['instructions'])

            # increment survey row when we're done adding all translations
            survey_row_counter += 1

            # then, finally, add all choice options (with all translations)
            if has_choices:
                values_added = []
                for translation in question_data:
                    for option in translation['options']:
                        if option['value'] not in values_added:
                            # if we haven't added this option value yet, add it for all translations
                            choices_ws.cell(row=choices_row_counter, column=choices_columns['list_name'],
                                            value=safe_question_id)
                            choices_ws.cell(row=choices_row_counter, column=choices_columns['value'],
                                            value=option['value'])
                            for inner_translation in question_data:
                                for inner_option in inner_translation['options']:
                                    if inner_option['value'] == option['value']:
                                        if (inner_translation['language'] and inner_translation['language'] !=
                                                primary_language):
                                            label_column = 'label' + f":{inner_translation['language']}"
                                        else:
                                            label_column = 'label'
                                        choices_ws.cell(row=choices_row_counter, column=choices_columns[label_column],
                                                        value=inner_option['label'])
                            choices_row_counter += 1
                            values_added.append(option['value'])

        # close the module's group
        survey_ws.cell(row=survey_row_counter, column=survey_columns['type'], value='end group')
        survey_ws.cell(row=survey_row_counter, column=survey_columns['name'], value=safe_module_name)
        survey_row_counter += 1

    # Save the workbook to the specified output file
    wb.save(output_file)



#  Copyright (c) 2024 Higher Bar AI, PBC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from setuptools import setup

with open('README.rst') as file:
    readme = file.read()

setup(
    name='surveyeval',
    version='0.1.11',
    packages=['surveyeval'],
    python_requires='>=3.10',
    install_requires=[
        'tiktoken',
        'openai~=1.29.0',
        'langchain~=0.1.15',
        'langchain-openai~=0.1.6',
        'langchain-community~=0.0.17',
        'overrides~=7.3.1',
    ],
    package_data={
        'surveyeval': ['resources/*'], # include resource files in package
    },
    package_dir={'': 'src'},
    url='https://github.com/higherbar-ai/survey-eval',
    project_urls={'Documentation': 'https://surveyeval.readthedocs.io/'},
    license='Apache 2.0',
    author='Christopher Robert',
    author_email='crobert@higherbar.ai',
    description='A toolkit for survey evaluation',
    long_description=readme
)
